"""
SixDesk Troubleshooting tools
-----------------------------

Some useful functions to troubleshoot the SixDesk output.
"""
import logging
from pathlib import Path

from pylhc_submitter.autosix import get_jobs_and_values
from pylhc_submitter.constants.autosix import (
    get_stagefile_path,
    get_track_path,
    get_workspace_path,
    SIXTRACK_INPUT_CHECK_FILES,
    SIXTRACK_OUTPUT_FILES,
    get_database_path,
)
from pylhc_submitter.sixdesk_tools.stages import STAGE_ORDER

LOG = logging.getLogger(__name__)


# Stages -----------------------------------------------------------------------


def get_last_stage(jobname, basedir):
    """Get the last run stage of job `jobname`."""
    stage_file = get_stagefile_path(jobname, basedir)
    last_stage = stage_file.read_text().strip("\n").split("\n")[-1]
    return STAGE_ORDER[last_stage]


# Set Stages ---

def set_stages_for_setup(basedir: Path, stage_name: str, jobid_mask: str, replace_dict: dict):
    """ Sets the last run stage for all jobs from given job-setups. """
    jobs, _ = get_jobs_and_values(jobid_mask, **replace_dict)
    for job in jobs:
        LOG.info(f"Setting stage to {stage_name} in {job}")
        set_stages(job, basedir, stage_name)


def set_stages(jobname: str, basedir: Path, stage_name: str):
    """Sets the last run stage of all given jobs to `stage_name`."""
    if stage_name not in STAGE_ORDER.keys():
        raise ValueError(f"Unknown stage '{stage_name}'")
    new_stage = STAGE_ORDER[stage_name]

    stages = []
    for stage in STAGE_ORDER.keys():
        stages.append(stage.name)
        if stage == new_stage:
            break

    stage_file = get_stagefile_path(jobname, basedir)
    stage_file.write_text("\n".join(stages))  # overwrites old file


def skip_stages(jobname: str, basedir: Path, stage_name: str):
    """Skip stages until `stagename`, i.e. similar to `set_stages` but only if
    the stage hasn't been reached yet. Inverse to `reset_stages`"""
    if stage_name not in STAGE_ORDER.keys():
        raise ValueError(f"Unknown stage '{stage_name}'")

    new_stage = STAGE_ORDER[stage_name]
    last_stage = get_last_stage(jobname, basedir)
    if last_stage < new_stage:
        LOG.info(
            f"Skipping stage form {last_stage.name} to {new_stage.name} in {jobname}"
        )
        set_stages(jobname, basedir, stage_name)
    else:
        LOG.debug(f"Stage {last_stage.name} unchanged in {jobname}")


def reset_stages(jobname: str, basedir: Path, stage_name: str):
    """Reset stages until `stagename`, i.e. similar to `set_stages` but only if
    the stage has already been run. Inverse to `skip_stages`"""
    if stage_name not in STAGE_ORDER.keys():
        raise ValueError(f"Unknown stage '{stage_name}'")

    new_stage = STAGE_ORDER[stage_name]
    last_stage = get_last_stage(jobname, basedir)
    if last_stage > new_stage:
        LOG.info(f"Resetting stage from {last_stage.name} to {new_stage.name} in {jobname}")
        set_stages(jobname, basedir, stage_name)
    else:
        LOG.debug(f"Stage {last_stage.name} unchanged in {jobname}")


# Check Stages ---

def check_stages_for_setup(basedir: Path, stage_name: str, jobid_mask: str, replace_dict: dict):
    """ Check the last run stage for all jobs from given job-setups. """
    jobs, _ = get_jobs_and_values(jobid_mask, **replace_dict)
    for job in jobs:
        check_last_stage(job, basedir)


def check_last_stage(jobname: str, basedir: Path):
    """Logs names of all last run stages for given jobs."""
    last_stage = get_last_stage(jobname, basedir)
    LOG.info(f"'{jobname}' at stage '{last_stage.name}'")


# Complete check for failure ---------------------------------------------------


def find_obviously_failed_sixtrack_submissions(basedir: Path):
    """Checks in jobs in `track` whether the directory structure seems to be created
    and if there is output data. This checks only the first directories found,
    to speed up this process. For a more precise scan see check_sixtrack_output_data.
    """
    jobs = []
    for job in get_all_jobs_in_base(basedir):
        try:
            LOG.debug(str(job))
            track = get_track_path(jobname=job, basedir=basedir)
            first_seed = get_first_dir(track) / "simul"
            tunes = get_first_dir(first_seed)
            amp = get_first_dir(tunes, "*_*")
            turns = get_first_dir(amp, "e*")
            angle = get_first_dir(turns, ".*")

            file_names = [f.name for f in angle.glob("*")]
            out_files_present = [f for f in SIXTRACK_OUTPUT_FILES if f in file_names]
            if not len(out_files_present):
                raise OSError(str(get_workspace_path(jobname=job, basedir=basedir)))
        except OSError as e:
            LOG.error(
                f"{e.args[0]} (stage: {get_last_stage(jobname=job, basedir=basedir).name})"
            )
            jobs.append(str(job))
    return jobs


def check_sixtrack_output_data(jobname: str, basedir: Path):
    """Presence checks for SixDesk tracking output data.

    This checks recursively all directories in `track`.
    Will be busy for a while.
    """
    track_path = get_track_path(jobname, basedir)
    seed_dirs = list(track_path.glob("[0-9]")) + list(track_path.glob("[0-9][0-9]"))
    if not len(seed_dirs):
        raise OSError(f"No seed-dirs present in {str(track_path)}.")

    for seed_dir in seed_dirs:
        if not seed_dir.is_dir():
            continue

        simul_path = seed_dir / "simul"
        tunes_dirs = list(simul_path.glob("*"))
        if not len(tunes_dirs):
            raise OSError(f"No tunes-dirs present in {str(seed_dir)}.")

        for tunes_dir in tunes_dirs:
            if not tunes_dir.is_dir():
                continue

            amp_dirs = list(tunes_dir.glob("*_*"))
            if not len(amp_dirs):
                raise OSError(f"No amplitude-dirs present in {str(tunes_dir)}.")

            for amp_dir in amp_dirs:
                if not amp_dir.is_dir():
                    continue

                turns_dirs = list(amp_dir.glob("*"))
                if not len(turns_dirs):
                    raise OSError(f"No turns-dirs present in {str(amp_dir)}.")

                for turn_dir in turns_dirs:
                    if not turn_dir.is_dir():
                        continue

                    angle_dirs = list(turn_dir.glob(".*"))
                    if not len(angle_dirs):
                        raise OSError(f"No angle-dirs present in {str(turn_dir)}.")

                    for angle_dir in angle_dirs:
                        if not angle_dir.is_dir():
                            continue

                        htcondor_files = list(angle_dir.glob("htcondor.*"))
                        if len(htcondor_files) != 3:
                            raise OSError(f"Not all htcondor files present in {str(angle_dir)}.")

                        file_names = [f.name for f in angle_dir.glob("*")]
                        in_files_present = [f for f in SIXTRACK_INPUT_CHECK_FILES if f in file_names]
                        if len(in_files_present):
                            raise OSError(
                                f"The files '{in_files_present}' are found in {str(angle_dir)},"
                                "yet they should have been deleted after tracking."
                            )

                        out_files_present = [f for f in SIXTRACK_OUTPUT_FILES if f in file_names]
                        if not len(out_files_present):
                            raise OSError(
                                f"None of the expected output files '{SIXTRACK_OUTPUT_FILES}' "
                                f"are present in {str(angle_dir)}"
                            )


# Long Database Names Hack -----------------------------------------------------


def create_database_symlink(jobname: str, basedir: Path):
    db_path = get_database_path(jobname, basedir)
    if db_path.exists():
        LOG.debug(f"Database already exists in {jobname}.")
        return

    real_db_path = db_path.parent / "my.db"
    real_db_path.touch()

    db_path.symlink_to(real_db_path)
    LOG.info(f"Crated database link in {jobname}.")


def move_database_symlink(jobname: str, basedir: Path):
    db_path = get_database_path(jobname, basedir)
    real_db_path = db_path.parent / "my.db"
    if real_db_path.exists():
        real_db_path.rename(db_path)
        LOG.info(f"Renamed database to its proper name in {jobname}.")


# Helper -----------------------------------------------------------------------


def for_all_jobs(func: callable, basedir: Path, *args, **kwargs):
    """Do function for all jobs in basedir."""
    for job in get_all_jobs_in_base(basedir):
        func(job, basedir, *args, **kwargs)


def get_all_jobs_in_base(basedir):
    """Returns all job-names in the sixdeskbase dir."""
    return [f.name.replace("workspace-", "") for f in basedir.glob("workspace-*")]


def get_first_dir(cwd: Path, glob: str = "*"):
    """Return first directory of pattern `glob`."""
    for d in cwd.glob(glob):
        if d.is_dir():
            return d
    raise OSError(str(cwd))




"""
Job Submitter
-------------

Collections of constants and paths used in the job-submitter.
"""
from pylhc_submitter.constants.external_paths import MADX_BIN, PYTHON2_BIN, PYTHON3_BIN

JOBSUMMARY_FILE = "Jobs.tfs"
JOBDIRECTORY_PREFIX = "Job"
CONFIG_FILE = "config.ini"

SCRIPT_EXTENSIONS = {
    "madx": ".madx",
    "python3": ".py",
    "python2": ".py",
}

EXECUTEABLEPATH = {
    "madx": MADX_BIN,
    "python3": PYTHON3_BIN,
    "python2": PYTHON2_BIN,
}


COLUMN_JOBID = "JobId"
COLUMN_SHELL_SCRIPT = "ShellScript"
COLUMN_JOB_DIRECTORY = "JobDirectory"
COLUMN_DEST_DIRECTORY = "DestDirectory"
COLUMN_JOB_FILE = "JobFile"

NON_PARAMETER_COLUMNS = (COLUMN_SHELL_SCRIPT, COLUMN_JOB_DIRECTORY, COLUMN_JOB_FILE, COLUMN_DEST_DIRECTORY)


"""
Post Process DA
----------------------------------

Tools to process data after sixdb has calculated the
da. Includes functions for extracting data from database
as well as plotting of DA polar plots.
"""
import logging
from pathlib import Path
from typing import Any, Tuple, Iterable

import numpy as np
import pandas as pd
from generic_parser import DotDict
from matplotlib import pyplot as plt
from matplotlib import rcParams, lines as mlines
from scipy.interpolate import interp1d
from tfs import TfsDataFrame, write_tfs

from pylhc_submitter.constants.autosix import (
    get_tfs_da_path,
    get_tfs_da_seed_stats_path,
    get_tfs_da_angle_stats_path,
    get_autosix_results_path,
    HEADER_NTOTAL,
    HEADER_INFO,
    HEADER_HINT,
    MEAN,
    STD,
    MIN,
    MAX,
    N,
    SEED,
    ANGLE,
    ALOST1,
    ALOST2,
    AMP,
)
from pylhc_submitter.sixdesk_tools.extract_data_from_db import extract_da_data

LOG = logging.getLogger(__name__)

DA_COLUMNS = (ALOST1, ALOST2)
INFO = (
    "Statistics over the N={n:d} {over:s} per {per:s}. "
    "The N-Columns indicate how many non-zero DA values were used."
)
HINT = "{param:s} {val:} is the respective value calculated over all other {param:s}s."
OVER_WHICH = {SEED: "angles", ANGLE: "seeds"}

# Fixed plot styles ---
COLOR_MEAN = "red"
COLOR_SEED = "grey"
COLOR_LIM = "black"
COLOR_FILL = "blue"
ALPHA_SEED = 0.5
ALPHA_FILL = 0.2
ALPHA_FILL_STD = 0.2


def post_process_da(jobname: str, basedir: Path):
    """ Post process the DA results into dataframes and DA plots. """
    LOG.info("Post-Processing Sixdesk Results.")
    df_da, df_angle, df_seed = create_da_tfs(jobname, basedir)
    create_polar_plots(jobname, basedir, df_da, df_angle)
    LOG.info("Post-Processing finished.")


# Data Analysis ----------------------------------------------------------------


def create_da_tfs(jobname: str, basedir: Path) -> Tuple[TfsDataFrame, TfsDataFrame, TfsDataFrame]:
    """ Extracts data from db into dataframes, and writes and returns them.

    Args:
        jobname (str): Name of the Job
        basedir (Path): SixDesk Basefolder Location
    """
    LOG.info("Gathering DA data into tfs-files.")
    df_da = extract_da_data(jobname, basedir)
    df_angle = _create_stats_df(df_da, ANGLE)
    df_seed = _create_stats_df(df_da, SEED, global_index=0)

    write_tfs(get_tfs_da_path(jobname, basedir), df_da)
    write_tfs(get_tfs_da_angle_stats_path(jobname, basedir), df_angle, save_index=ANGLE)
    write_tfs(get_tfs_da_seed_stats_path(jobname, basedir), df_seed, save_index=SEED)
    return df_da, df_angle, df_seed


def _create_stats_df(df: pd.DataFrame, parameter: str, global_index: Any = None) -> TfsDataFrame:
    """ Calculates the stats over a given parameter.
    Note: Could be refactored to use `group_by`.

    Args:
        df (DataFrame): DataFrame containing the DA information over all seeds.
        parameter (str): The parameter over which we want to average, i.e.
                         SEED or ANGLE.
        global_index (Any): identifier to use as a global index, i.e. the statistics
                            over all entries are stored here. (e.g. '0' for SEEDs)

    """
    operation_map = DotDict({MEAN: np.mean, STD: np.std, MIN: np.min, MAX: np.max})

    pre_index = [] if global_index is None else [global_index]
    index = sorted(set(df[parameter]))
    n_total = sum(df[parameter] == index[0])

    df_stats = TfsDataFrame(
        index=pre_index + index,
        columns=[f"{fun}{al}" for al in DA_COLUMNS for fun in list(operation_map.keys()) + [N]],
    )
    df_stats.headers[HEADER_INFO] = INFO.format(
        over=OVER_WHICH[parameter], per=parameter.lower(), n=n_total
    )
    df_stats.headers[HEADER_NTOTAL] = n_total

    for col_da in DA_COLUMNS:
        for idx in index:
            mask = (df[parameter] == idx) & (df[col_da] != 0)
            df_stats.loc[idx, f"{N}{col_da}"] = sum(mask)
            for name, operation in operation_map.items():
                df_stats.loc[idx, f"{name}{col_da}"] = operation(df.loc[mask, col_da])
            for name, operation in operation_map.get_subdict([MIN, MAX]).items():
                df_stats.loc[idx, f"{name}{AMP}"] = operation(df.loc[mask, f"{name}{AMP}"])

        if global_index is not None:
            # Note: could be done over df_stats for MEAN, MIN and MAX, but not STD
            mask = df[col_da] != 0
            df_stats.loc[global_index, f"{N}{col_da}"] = sum(mask)

            # Global MEAN, MIN, MAX Dynamic Aperture
            for name, operation in operation_map.get_subdict([MEAN, MIN, MAX, STD]).items():
                df_stats.loc[global_index, f"{name}{col_da}"] = operation(
                    df.loc[mask, col_da]
                )

            # Global MIN, MAX Amplitudes
            for name, operation in operation_map.get_subdict([MIN, MAX]).items():
                df_stats.loc[global_index, f"{name}{AMP}"] = operation(
                    df.loc[mask, f"{name}{AMP}"]  # min(MINA) and max(MAXA)
                )

            df_stats.headers[HEADER_HINT] = HINT.format(param=parameter, val=global_index)

    return df_stats


# Single Plots -----------------------------------------------------------------


def create_polar_plots(jobname: str, basedir: Path, df_da: TfsDataFrame, df_angles: TfsDataFrame):
    """ Plotting loop over da-methods and wrapper so save plots.

    Args:
        jobname (str): Name of the Job
        basedir (Path): SixDesk Basefolder Location
        df_da (TfsDataFrame): Full DA analysis result.
        df_angles (TfsDataFrame): Dataframe with the statistics (min, max, mean) per angle
    """
    LOG.info("Creating Polar Plots.")
    outdir_path = get_autosix_results_path(jobname, basedir)
    for da_col in DA_COLUMNS:
        fig = plot_polar(df_angles, da_col, jobname, df_da)
        fig.tight_layout(), fig.tight_layout()
        fig.savefig(outdir_path / fig.canvas.get_default_filename())

    # plt.show()


def plot_polar(
    df_angles: TfsDataFrame,
    da_col: str = ALOST2,
    jobname: str = "",
    df_da: TfsDataFrame = None,
    **kwargs,
) -> plt.Figure:
    """Create Polar Plot for DA analysis data.

    Keyword arguments are all optional.

    Args:
        df_angles (TfsDataFrame): Dataframe with the statistics (min, max, mean) per angle
        da_col (str): DA-Column name from sixdesk analysis to be used , e.g. ``ALOST2``.
                      (optional, default: ``ALOST2``)
        jobname (str): Name of the job. Used in window title only (optional).
        df_da (TfsDataFrame): Full DA analysis result. If given, plots
                              the individual DA results per seed. (optional)

    Keyword Arguments:
        interpolated (bool): If true, uses interpolation to plot the lines curved
        fill (bool): If true, fills the area between min and max with light blue
        angle_ticks (Iterable[numeric]): Positions in degree of the angle ticks (and lines)
        amplitude ticks (Iterable[numeric]): Positions of the amplitude ticks.

    Returns:
        Figure of the polar plot.
    """
    interpolated: bool = kwargs.pop("interpolated", True)
    fill: bool = kwargs.pop("fill", df_da is None)
    angle_ticks: Iterable[np.numeric] = kwargs.pop("angle_ticks", None)
    amplitude_ticks: Iterable[np.numeric] = kwargs.pop("amplitude_ticks", None)

    if "lines.marker" not in kwargs:
        kwargs["lines.marker"] = "None"
    fig, ax = plt.subplots(nrows=1, ncols=1, subplot_kw={"projection": "polar"})
    fig.canvas.manager.set_window_title(f"{jobname} polar plot for {da_col}")

    angles = np.deg2rad(df_angles.index)
    da_min, da_mean, da_max, da_std = (
        df_angles[f"{name}{da_col}"] for name in (MIN, MEAN, MAX, STD)
    )

    seed_h, seed_l = _plot_seeds(ax, df_da, da_col, interpolated)
    if interpolated:
        mean_h, max_h = _plot_interpolated(ax, angles, da_min, da_mean, da_max, da_std, fill)
    else:
        mean_h, max_h = _plot_straight(ax, angles, da_min, da_mean, da_max, da_std, fill)

    ax.set_thetamin(0)
    ax.set_thetamax(90)
    ax.set_rlim([0, None])
    ax.set_xlabel(r"DA$_{x}~[\sigma_{nominal}]$", labelpad=15)
    ax.set_ylabel(r"DA$_{y}~[\sigma_{nominal}]$", labelpad=20)

    if angle_ticks is not None:
        ax.set_xticks(np.deg2rad(angle_ticks))

    if amplitude_ticks is not None:
        ax.set_yticks(amplitude_ticks)
    ax.tick_params(labelright=True, labelleft=True)

    ax.legend(
        loc="upper right",
        bbox_to_anchor=(0.9, 0.95),
        bbox_transform=fig.transFigure,
        # frameon=False,
        handles=seed_h + [mean_h, max_h],
        labels=seed_l + ["Mean DA", "Limits"],
        ncol=1,
    )

    return fig


def _plot_seeds(ax, df_da: TfsDataFrame, da_col: str, interpolated: bool) -> Tuple[list, list]:
    """Add the Seed lines to the polar plots, if df_da is given.

    Args:
        ax: Axes to plot in
        df_da: DataFrame with DA information
        da_col: Dynamic Aperture column (ALOST1 or ALOST2)
        interpolated: If true, the lines will be curved.

    Returns:
        Tuple of list of one line handle and a list of a single label
    """
    if df_da is not None:
        seed_h = None
        for seed in sorted(set(df_da[SEED])):
            seed_mask = df_da[SEED] == seed
            angles = np.deg2rad(df_da.loc[seed_mask, ANGLE])
            da_data = df_da.loc[seed_mask, da_col]
            da_data.loc[da_data == 0] = np.NaN
            if interpolated:
                seed_h, _, _ = _interpolated_line(
                    ax,
                    angles,
                    da_data,
                    c=COLOR_SEED,
                    ls="-",
                    label=f"Seed {seed:d}",
                    alpha=ALPHA_SEED,
                )
            else:
                (seed_h,) = ax.plot(
                    angles, da_data, c=COLOR_SEED, ls="-", label=f"Seed {seed:d}", alpha=ALPHA_SEED
                )
        return [seed_h], ["DA per Seed"]
    return [], []


def _plot_interpolated(ax, angles, da_min, da_mean, da_max, da_std, fill):
    """Plot interpolated DA lines and areas."""
    _, _, ip_min = _interpolated_line(ax, angles, da_min, c=COLOR_LIM, ls="--", label="Minimum DA")
    max_h, ip_x, ip_max = _interpolated_line(
        ax, angles, da_max, c=COLOR_LIM, ls="--", label="Maximum DA"
    )
    if fill:
        ax.fill_between(ip_x, ip_min, ip_max, color=COLOR_FILL, alpha=ALPHA_FILL)
        _, ip_std_min = _interpolated_coords(angles, da_mean - da_std)
        _, ip_std_max = _interpolated_coords(angles, da_mean + da_std)
        ax.fill_between(ip_x, ip_std_min, ip_std_max, color=COLOR_FILL, alpha=ALPHA_FILL_STD)

    mean_h, _, _ = _interpolated_line(ax, angles, da_mean, c=COLOR_MEAN, ls="-", label="Mean DA")
    return mean_h, max_h


def _plot_straight(ax, angles, da_min, da_mean, da_max, da_std, fill):
    """Plot straight DA lines and areas."""
    (_,) = ax.plot(angles, da_min, c=COLOR_LIM, ls="--", label="Minimum DA")
    (max_h,) = ax.plot(angles, da_max, c=COLOR_LIM, ls="--", label="Maximum DA")
    if fill:
        ax.fill_between(
            angles,
            da_min.astype(float),
            da_max.astype(float),  # weird conversion to obj otherwise
            color=COLOR_FILL,
            alpha=ALPHA_FILL,
        )
        ax.fill_between(
            angles,
            (da_mean - da_std).astype(float),
            (da_mean + da_std).astype(float),
            color=COLOR_FILL,
            alpha=ALPHA_FILL_STD,
        )
    (mean_h,) = ax.plot(angles, da_mean, c=COLOR_MEAN, ls="-", label="Mean DA")
    return mean_h, max_h


def _interpolated_line(ax, x, y, npoints: int = 100, **kwargs):
    """Plot a line that interpolates linearly between points.
    Useful for polar plots with sparse points."""
    ls = kwargs.pop("linestyle", kwargs.pop("ls", rcParams["lines.linestyle"]))
    marker = kwargs.pop("marker", rcParams["lines.marker"])
    label = kwargs.pop("label")

    ip_x, ip_y = _interpolated_coords(x, y, npoints)
    (line_h,) = ax.plot(ip_x, ip_y, marker="None", ls=ls, label=f"_{label}_line", **kwargs)

    if marker.lower() not in ["none", ""]:
        ax.plot(x, y, ls="None", marker=marker, label=f"_{label}_markers", **kwargs)

    # fake handle for legend
    handle = mlines.Line2D([], [], color=line_h.get_color(), ls=ls, marker=marker, label=label)
    return handle, ip_x, ip_y


def _interpolated_coords(x, y, npoints: int = 100):
    """ Do linear interpolation between points. """
    ip_x = np.linspace(min(x), max(x), npoints)
    ip_y = interp1d(x, y)(ip_x)
    return ip_x, ip_y



"""
External Paths
--------------

Specific constants relating to external paths to be used,
to help with consistency.
"""
from pathlib import Path

AFS_CERN: Path = Path("/", "afs", "cern.ch")
LINTRACK: Path = AFS_CERN / "eng" / "sl" / "lintrack"

# Binary Files -----------------------------------------------------------------
MADX_BIN: Path = AFS_CERN / "user" / "m" / "mad" / "bin" / "madx"
PYTHON3_BIN: Path = LINTRACK / "omc_python3" / "bin" / "python"
PYTHON2_BIN: Path = LINTRACK / "miniconda2" / "bin" / "python"
SIXDESK_UTILS: Path = AFS_CERN / "project" / "sixtrack" / "SixDesk_utilities" / "pro"



"""
pylhc-submitter
~~~~~~~~~~~~~~~~

pylhc-submitter contains scripts to simplify the creation and submission of jobs to HTCondor at CERN.

:copyright: pyLHC/OMC-Team working group.
:license: MIT, see the LICENSE.md file for details.
"""
__title__ = "pylhc_submitter"
__description__ = "pylhc-submitter contains scripts to simplify the creation and submission of jobs to HTCondor at CERN"
__url__ = "https://github.com/pylhc/submitter"
__version__ = "2.0.2"
__author__ = "pylhc"
__author_email__ = "pylhc@github.com"
__license__ = "MIT"

__all__ = [__version__]



"""
Job Submitter
-------------


The ``job_submitter`` allows to execute a parametric study using a script mask and a `dictionary` of parameters to replace in this mask, from the command line.
These parameters must be present in the given mask in the ``%(PARAMETER)s`` format (other types apart from string are also allowed).

The type of script and executable is freely choosable, but defaults to ``madx``, for which this submitter was originally written.
When submitting to ``HTCondor``, data to be transferred back to the working directory must be written in a sub-folder defined by ``job_output_directory`` which defaults to **Outputdata**.

This script also allows to check if all ``HTCondor`` jobs finished successfully, for resubmissions with a different parameter grid, and for local execution.
A **Jobs.tfs** file is created in the working directory containing the Job Id, parameter per job
and job directory for further post processing.

For additional information and guides, see the `Job Submitter page
<https://pylhc.github.io/packages/pylhcsubmitter/job_submitter.html>`_ in the ``OMC`` documentation site.


*--Required--*

- **mask** *(PathOrStr)*:

    Program mask to use


- **replace_dict** *(DictAsString)*:

    Dict containing the str to replace as keys and values a list of
    parameters to replace


- **working_directory** *(PathOrStr)*:

    Directory where data should be put


*--Optional--*

- **append_jobs**:

    Flag to rerun job with finer/wider grid, already existing points will
    not be reexecuted.

    action: ``store_true``


- **check_files** *(str)*:

    List of files/file-name-masks expected to be in the 'job_output_dir'
    after a successful job (for appending/resuming). Uses the 'glob'
    function, so unix-wildcards (*) are allowed. If not given, only the
    presence of the folder itself is checked.


- **dryrun**:

    Flag to only prepare folders and scripts, but does not start/submit
    jobs. Together with `resume_jobs` this can be use to check which jobs
    succeeded and which failed.

    action: ``store_true``


- **executable** *(PathOrStr)*:

    Path to executable or job-type (of ['madx', 'python3', 'python2']) to
    use.

    default: ``madx``


- **htc_arguments** *(DictAsString)*:

    Additional arguments for htcondor, as Dict-String. For AccountingGroup
    please use 'accounting_group'. 'max_retries' and 'notification' have
    defaults (if not given). Others are just passed on.

    default: ``{}``


- **job_output_dir** *(str)*:

    The name of the output dir of the job. (Make sure your script puts its
    data there!)

    default: ``Outputdata``


- **jobflavour** *(str)*:

    Jobflavour to give rough estimate of runtime of one job

    choices: ``('espresso', 'microcentury', 'longlunch', 'workday', 'tomorrow', 'testmatch', 'nextweek')``

    default: ``workday``


- **jobid_mask** *(str)*:

    Mask to name jobs from replace_dict


- **num_processes** *(int)*:

    Number of processes to be used if run locally

    default: ``4``


- **output_destination** *(PathOrStr)*:

    Directory to copy the output of the jobs to, sorted into folders per job.
    Can be on EOS, preferrably via EOS-URI format ('root://eosuser.cern.ch//eos/...').


- **resume_jobs**:

    Only do jobs that did not work.

    action: ``store_true``


- **run_local**:

    Flag to run the jobs on the local machine. Not suggested.

    action: ``store_true``


- **script_arguments** *(DictAsString)*:

    Additional arguments to pass to the script, as dict in key-value pairs
    ('--' need to be included in the keys).

    default: ``{}``


- **script_extension** *(str)*:

    New extension for the scripts created from the masks. This is inferred
    automatically for ['madx', 'python3', 'python2']. Otherwise not
    changed.


- **ssh** *(str)*:

    Run htcondor from this machine via ssh (needs access to the
    `working_directory`)


"""
import logging
import sys
from dataclasses import fields
from pathlib import Path

from generic_parser import EntryPointParameters, entrypoint
from generic_parser.entry_datatypes import DictAsString
from generic_parser.tools import print_dict_tree

from pylhc_submitter.constants.htcondor import JOBFLAVOURS
from pylhc_submitter.constants.job_submitter import EXECUTEABLEPATH, SCRIPT_EXTENSIONS
from pylhc_submitter.submitter.iotools import CreationOpts, create_jobs, is_eos_uri, print_stats
from pylhc_submitter.submitter.mask import (check_percentage_signs_in_mask,
                                            find_named_variables_in_mask, is_mask_file)
from pylhc_submitter.submitter.runners import RunnerOpts, run_jobs
from pylhc_submitter.utils.iotools import (PathOrStr, keys_to_path, make_replace_entries_iterable,
                                           save_config)
from pylhc_submitter.utils.logging_tools import log_setup

LOG = logging.getLogger(__name__)


try:
    import htcondor
except ImportError:
    platform = "macOS" if sys.platform == "darwin" else "windows"
    LOG.warning(
        f"htcondor python bindings are linux-only. You can still use job_submitter on {platform}, "
        "but only for local runs."
    )
    htcondor = None


def get_params():
    params = EntryPointParameters()
    params.add_parameter(
        name="mask",
        type=PathOrStr,
        required=True,
        help="Program mask to use",
    )
    params.add_parameter(
        name="working_directory",
        type=PathOrStr,
        required=True,
        help="Directory where data should be put",
    )
    params.add_parameter(
        name="executable",
        default="madx",
        type=PathOrStr,
        help=(
            "Path to executable or job-type " f"(of {str(list(EXECUTEABLEPATH.keys()))}) to use."
        ),
    )
    params.add_parameter(
        name="jobflavour",
        type=str,
        choices=JOBFLAVOURS,
        default="workday",
        help="Jobflavour to give rough estimate of runtime of one job ",
    )
    params.add_parameter(
        name="run_local",
        action="store_true",
        help="Flag to run the jobs on the local machine. Not suggested.",
    )
    params.add_parameter(
        name="resume_jobs",
        action="store_true",
        help="Only do jobs that did not work.",
    )
    params.add_parameter(
        name="append_jobs",
        action="store_true",
        help=(
            "Flag to rerun job with finer/wider grid, already existing points will not be "
            "reexecuted."
        ),
    )
    params.add_parameter(
        name="dryrun",
        action="store_true",
        help=(
            "Flag to only prepare folders and scripts, but does not start/submit jobs. "
            "Together with `resume_jobs` this can be use to check which jobs "
            "succeeded and which failed."
        ),
    )
    params.add_parameter(
        name="replace_dict",
        help=(
            "Dict containing the str to replace as keys and values a list of parameters to "
            "replace"
        ),
        type=DictAsString,
        required=True,
    )
    params.add_parameter(
        name="script_arguments",
        help=(
            "Additional arguments to pass to the script, as dict in key-value pairs "
            "('--' need to be included in the keys)."
        ),
        type=DictAsString,
        default={},
    )
    params.add_parameter(
        name="script_extension",
        help=(
            "New extension for the scripts created from the masks. This is inferred "
            f"automatically for {str(list(SCRIPT_EXTENSIONS.keys()))}. Otherwise not changed."
        ),
        type=str,
    )
    params.add_parameter(
        name="num_processes",
        help="Number of processes to be used if run locally",
        type=int,
        default=4,
    )
    params.add_parameter(
        name="check_files",
        help=(
            "List of files/file-name-masks expected to be in the "
            "'job_output_dir' after a successful job "
            "(for appending/resuming). Uses the 'glob' function, so "
            "unix-wildcards (*) are allowed. If not given, only the "
            "presence of the folder itself is checked."
        ),
        type=str,
        nargs="+",
    )
    params.add_parameter(
        name="jobid_mask",
        help="Mask to name jobs from replace_dict",
        type=str,
    )
    params.add_parameter(
        name="job_output_dir",
        help="The name of the output dir of the job. (Make sure your script puts its data there!)",
        type=str,
        default="Outputdata",
    )
    params.add_parameter(
        name="output_destination",
        help="Directory to copy the output of the jobs to, sorted into folders per job. "
             "Can be on EOS, preferrably via EOS-URI format ('root://eosuser.cern.ch//eos/...').",
        type=PathOrStr,
    )
    params.add_parameter(
        name="htc_arguments",
        help=(
            "Additional arguments for htcondor, as Dict-String. "
            "For AccountingGroup please use 'accounting_group'. "
            "'max_retries' and 'notification' have defaults (if not given). "
            "Others are just passed on. "
        ),
        type=DictAsString,
        default={},
    )
    params.add_parameter(
        name="ssh",
        help="Run htcondor from this machine via ssh (needs access to the `working_directory`)",
        type=str,
    )

    return params


@entrypoint(get_params(), strict=True)
def main(opt):
    if not opt.run_local:
        LOG.info("Starting HTCondor Job-submitter.")
        _check_htcondor_presence()
    else:
        LOG.info("Starting Job-submitter.")

    save_config(Path(opt.working_directory), opt, "job_submitter")
    creation_opt, runner_opt = check_opts(opt)

    job_df, dropped_jobs = create_jobs(creation_opt)

    run_jobs(job_df, runner_opt)

    print_stats(job_df.index, dropped_jobs)


def check_opts(opt):
    """ Checks options and sorts them into job-creation and running parameters. """
    LOG.debug("Checking options.")
    if opt.resume_jobs and opt.append_jobs:
        raise ValueError("Select either Resume jobs or Append jobs")

    # Paths ---
    opt = keys_to_path(opt, "working_directory", "executable")

    if str(opt.executable) in EXECUTEABLEPATH.keys():
        opt.executable = str(opt.executable)

    if is_mask_file(opt.mask):
        mask_content = Path(opt.mask).read_text()  # checks that mask and dir are there
        opt.mask = Path(opt.mask)
    else:
        mask_content = opt.mask
    
    if is_eos_uri(opt.output_destination) and not ("://" in opt.output_destination and "//eos/" in opt.output_destination):
        raise ValueError(
            "The 'output_destination' is an EOS-URI but missing '://' or '//eos' (double slashes?). "
        )
        

    # Replace dict ---
    dict_keys = set(opt.replace_dict.keys())
    mask_keys = find_named_variables_in_mask(mask_content)
    not_in_mask = dict_keys - mask_keys
    not_in_dict = mask_keys - dict_keys

    if len(not_in_dict):
        raise KeyError(
            "The following keys in the mask were not found in the given replace_dict: "
            f"{str(not_in_dict).strip('{}')}"
        )

    if len(not_in_mask):
        LOG.warning(
            "The following replace_dict keys were not found in the given mask: "
            f"{str(not_in_mask).strip('{}')}"
        )

        # remove all keys which are not present in mask (otherwise unnecessary jobs)
        [opt.replace_dict.pop(key) for key in not_in_mask]
        if len(opt.replace_dict) == 0:
            raise KeyError("Empty replace-dictionary")
    check_percentage_signs_in_mask(mask_content)

    print_dict_tree(opt, name="Input parameter", print_fun=LOG.debug)
    opt.replace_dict = make_replace_entries_iterable(opt.replace_dict)
    
    # Create new classes
    opt.output_dir = opt.job_output_dir  # renaming

    creation = CreationOpts(**{f.name: opt[f.name] for f in fields(CreationOpts)})
    runner = RunnerOpts(**{f.name: opt[f.name] for f in fields(RunnerOpts)})
    runner.output_dir = None if opt.output_destination else opt.output_dir
    return creation, runner


def _check_htcondor_presence() -> None:
    """ Raises an error if htcondor is not installed. """
    if htcondor is None:
        raise EnvironmentError("htcondor bindings are necessary to run this module.")


# Script Mode ------------------------------------------------------------------


if __name__ == "__main__":
    log_setup()
    main()



"""
Environment Tools
-----------------

Some tools that depend on the environment of the run.
"""
import sys


def on_windows() -> bool:
    return sys.platform.startswith("win")


def on_linux() -> bool:
    return sys.platform == "linux"



"""
Stages
------

In this module the stages are organized.

"""
import logging
import re
from abc import ABC, abstractmethod, ABCMeta

from generic_parser import DotDict

from pylhc_submitter.constants.autosix import get_stagefile_path, StageSkip, StageStop, AutoSixEnvironment
from pylhc_submitter.sixdesk_tools.create_workspace import (
    create_job, init_workspace, fix_pythonfile_call,
    remove_twiss_fail_check, set_max_materialize
)
from pylhc_submitter.sixdesk_tools.post_process_da import post_process_da
from pylhc_submitter.sixdesk_tools.submit import (
    submit_mask, check_sixtrack_input, submit_sixtrack,
    check_sixtrack_output, sixdb_load, sixdb_cmd
)

LOG = logging.getLogger(__name__)

# Overwritten in StageMeta below and actual classes inserted
STAGE_ORDER = DotDict({
    "create_job": None,
    "initialize_workspace": None,
    "submit_mask": None,
    "check_input": None,
    "submit_sixtrack": None,
    "check_sixtrack_output": None,
    "sixdb_load": None,
    "sixdb_cmd": None,
    "post_process": None,
    "final": None,
})


class StageMeta(ABCMeta):
    """ Dynamically generate name and value from STAGE_ORDER """
    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # convert CamelCase to snake_case (https://stackoverflow.com/a/1176023/5609590)
        cls.name = re.sub(r'(?<!^)(?=[A-Z])', '_', cls.__name__).lower()

        if cls.name == "stage":
            # Base Class
            cls.value = None
            return

        # set value according to list (so we can compare order)
        cls.value = list(STAGE_ORDER.keys()).index(cls.name)

        # set class as entry in STAGE_ORDER
        STAGE_ORDER[cls.name] = cls
    
    def __str__(cls):
        return cls.name

    def __int__(cls):
        return cls.value

    def __add__(cls, other):
        new_idx = int(cls) + int(other)
        return list(STAGE_ORDER.values())[new_idx]

    def __sub__(cls, other):
        new_idx = int(cls) - int(other)
        if new_idx < 0:
            raise IndexError("Requested Stage order is negative. Stage not found.")
        return list(STAGE_ORDER.values())[new_idx]
    
    def __gt__(cls, other):
        try:
            return int(cls) > int(other)
        except TypeError:
            return False

    def __ge__(cls, other):
        try:
            return int(cls) >= int(other)
        except TypeError:
            return False

    def __lt__(cls, other):
        try:
            return int(cls) < int(other)
        except TypeError:
            return False

    def __le__(cls, other):
        try:
            return int(cls) <= int(other)
        except TypeError:
            return False

    def __eq__(cls, other):
        try:
            return int(cls) == int(other)
        except TypeError:
            return False

    def __hash__(cls):
        return hash((cls.value, cls.name))


class Stage(ABC, metaclass=StageMeta):
    """ Abstract Base Class for all Stages. Also provides the basic methods.
    The stages themselves only need to implement the _run() method. """
    @staticmethod
    def run_all_stages(jobname, jobargs, env):
        """ Run all stages in order. """
        LOG.info(f"vv---------------- Job {jobname} -------------------vv")
        for stage_name, stage_class in STAGE_ORDER.items():
            stage = stage_class(jobname, jobargs, env)
            try:
                stage.run()
            except StageSkip as e:
                if str(e):
                    LOG.error(e)
                # break  # stop here or always run to the end and show all skipped stages
            except StageStop:
                LOG.info(f"Stopping after Stage '{stage!s}' as the submitted jobs will now run. "
                         f"Check `condor_q` for their progress and restart autosix when they are done.")
                break
        LOG.info(f"^^---------------- Job {jobname} -------------------^^")

    def __init__(self, jobname: str, jobargs: dict, env: AutoSixEnvironment):
        self.jobname = jobname
        self.env = env
        self.jobargs = jobargs

        # Helper ---
        self.basedir = env.working_directory
        self.max_stage = env.max_stage
        self.stage_file = get_stagefile_path(self.jobname, self.basedir)

    def __str__(self):
        return self.name

    def __int__(self):
        return self.value

    def __sub__(self, other):
        return StageMeta.__sub__(self.__class__, other)(self.jobname, self.jobargs, self.env)

    def __add__(self, other):
        return StageMeta.__add__(self.__class__, other)(self.jobname, self.jobargs, self.env)

    def __gt__(self, other):
        return int(self) > int(other)

    def __ge__(self, other):
        return int(self) >= int(other)

    def __lt__(self, other):
        return int(self) < int(other)

    def __le__(self, other):
        return int(self) <= int(other)

    def __eq__(self, other):
        return int(self) == int(other)

    def should_run_stage(self):
        """ Checks if the stage should be run. """
        if not self.stage_file.exists():
            if self == 0:
                return True
            else:
                LOG.info(f"Stage '{self!s}' not run because previous stage(s) missing.")
                return False

        stage_file_txt = self.stage_file.read_text().split("\n")
        run_stages = [line.strip() for line in stage_file_txt if line.strip()]

        if self.name in run_stages:
            LOG.info(f"Stage '{self!s}' has already been run. Skipping.")
            return False

        if self == 0:
            return True

        # check if user requested a stop at a certain stage
        if (self.max_stage is not None) and (self > self.max_stage):
            LOG.info(f"Stage '{self!s}' would run after requested "
                     f"maximum stage '{self.max_stage!s}'. Skipping.")
            return False

        # check if last run stage is also the stage before current stage in stage order
        if run_stages[-1] == (self - 1).name:
            return True

        LOG.info(f"Stage '{self!s}' not run because previous stage(s) missing.")
        return False

    def stage_done(self):
        """ Append current stage name to stagefile. """
        with open(self.stage_file, "a+") as f:
            f.write(f"{self!s}\n")

    def run(self):
        """ Run the Stage. """
        if not self.should_run_stage():
            return

        try:
            self._run()
        except StageStop as e:
            # Stage indicates that it ran successfully,
            # but that there should be a stop in the loop.
            self.stage_done()
            raise e
        except StageSkip as e:
            # logged/handled outside
            raise e
        except Exception as e:
            # convert any exception to a StageSkip,
            # so the other jobs can continue running.
            LOG.exception(str(e))
            raise StageSkip(f"Stage {self!s} failed!") from e

        self.stage_done()

    @abstractmethod
    def _run(self):
        pass


# Actual Stages ----------------------------------------------------------------
# These Stages should actually have the function they call implemented directly
# and should be defined in the modules directly. But I like that you can read
# them one after another here and also see the command one would have to run if
# one would do this manually (which can be very helpful for fixing broken jobs).
# - jdilly 2020-08-04

class CreateJob(Stage):
    """
    create workspace
    > cd $basedir
    > /afs/cern.ch/project/sixtrack/SixDesk_utilities/pro/utilities/bash/set_env.sh -N workspace-$jobname

    write sixdeskenv, sysenv, filled mask (manual)
    """
    def _run(self):
        create_job(self.jobname, self.basedir,
                   executable=self.env.executable,
                   mask_text=self.env.mask_text,
                   sixdesk=self.env.sixdesk_directory,
                   ssh=self.env.ssh,
                   **self.jobargs)


class InitializeWorkspace(Stage):
    """
    initialize workspace
    > cd $basedir/workspace-$jobname/sixjobs
    > /afs/cern.ch/project/sixtrack/SixDesk_utilities/pro/utilities/bash/set_env.sh -s

    remove the twiss-fail check in sixtrack_input
    (manual)
    """
    def _run(self):
        if self.env.stop_workspace_init:
            LOG.info(
                f"Workspace creation for job {self.jobname} interrupted."
                " Check directory to manually adapt ``sixdeskenv``"
                " and ``sysenv``. Remove 'stop_workspace_init' from input"
                " parameters or set to 'False' to continue run."
            )
            raise StageSkip()

        init_workspace(self.jobname, self.basedir,
                       sixdesk=self.env.sixdesk_directory,
                       ssh=self.env.ssh)
        if self.env.apply_mad6t_hacks:
            fix_pythonfile_call(self.jobname, self.basedir)  # removes "<" in call
            remove_twiss_fail_check(self.jobname, self.basedir)  # removes 'grep twiss fail'


class SubmitMask(Stage):
    """
    submit for input generation
    > cd $basedir/workspace-$jobname/sixjobs
    > /afs/cern.ch/project/sixtrack/SixDesk_utilities/pro/utilities/bash/mad6t.sh -s
    """
    def _run(self):
        submit_mask(self.jobname, self.basedir,
                    sixdesk=self.env.sixdesk_directory,
                    ssh=self.env.ssh)
        raise StageStop()


class CheckInput(Stage):
    """
    Check if input files have been generated properly
    > cd $basedir/workspace-$jobname/sixjobs
    > /afs/cern.ch/project/sixtrack/SixDesk_utilities/pro/utilities/bash/mad6t.sh -c

    If not, and resubmit is active
    > /afs/cern.ch/project/sixtrack/SixDesk_utilities/pro/utilities/bash/mad6t.sh -w
    """
    def _run(self):
        check_sixtrack_input(self.jobname, self.basedir,
                             sixdesk=self.env.sixdesk_directory,
                             ssh=self.env.ssh,
                             resubmit=self.env.resubmit)


class SubmitSixtrack(Stage):
    """
    Generate simulation files (-g) and check if runnable (-c) and submit (-s) (-g -c -s == -a).
    > cd $basedir/workspace-$jobname/sixjobs
    > /afs/cern.ch/project/sixtrack/SixDesk_utilities/pro/utilities/bash/run_six.sh -a
    """
    def _run(self):
        # adds max_materialize to tracking sub-file template
        # might run into a race condition, this is why it's done here.
        # Also I assume we use the same value for all jobs anyway.
        set_max_materialize(self.env.sixdesk_directory, self.env.max_materialize)

        submit_sixtrack(self.jobname, self.basedir,
                        sixdesk=self.env.sixdesk_directory,
                        ssh=self.env.ssh,
                        python=self.env.python2)

        raise StageStop()


class CheckSixtrackOutput(Stage):
    """
    Checks sixtrack output via run_status. If this fails even though all
    jobs have finished on the scheduler, check the log-output (run_status
    messages are logged to debug).
    > cd $basedir/workspace-$jobname/sixjobs
    > /afs/cern.ch/project/sixtrack/SixDesk_utilities/pro/utilities/bash/run_status

    If not, and resubmit is active
    > cd $basedir/workspace-$jobname/sixjobs
    > /afs/cern.ch/project/sixtrack/SixDesk_utilities/pro/utilities/bash/run_six.sh -i
    """
    def _run(self):
        check_sixtrack_output(self.jobname, self.basedir,
                              sixdesk=self.env.sixdesk_directory,
                              ssh=self.env.ssh,
                              python=self.env.python2,
                              resubmit=self.env.resubmit)


class SixdbLoad(Stage):
    """
    Gather results into database via sixdb.
    > cd $basedir/workspace-$jobname/sixjobs
    > python3 /afs/cern.ch/project/sixtrack/SixDesk_utilities/pro/utilities/externals/SixDeskDB/sixdb . load_dir
    """
    def _run(self):
        sixdb_load(self.jobname, self.basedir,
                   sixdesk=self.env.sixdesk_directory,
                   ssh=self.env.ssh,
                   python=self.env.python3)


class SixdbCmd(Stage):
    """
    Analysise results in database via sixdb.
    > cd $basedir/workspace-$jobname/sixjobs
    > python3 /afs/cern.ch/project/sixtrack/SixDesk_utilities/pro/utilities/externals/SixDeskDB/sixdb $jobname da

    when fixed:
    > python3 /afs/cern.ch/project/sixtrack/SixDesk_utilities/pro/utilities/externals/SixDeskDB/sixdb $jobname da_vs_turns -turnstep 100 -outfile
    > python3 /afs/cern.ch/project/sixtrack/SixDesk_utilities/pro/utilities/externals/SixDeskDB/sixdb $jobname plot_da_vs_turns
    """
    def _run(self):
        sixdb_cmd(self.jobname, self.basedir,
                  sixdesk=self.env.sixdesk_directory,
                  ssh=self.env.ssh,
                  cmd=["da"],
                  python=self.env.python3)

        # da_vs_turns is broken at the moment (jdilly, 19.10.2020)
        # sixdb_cmd(jobname, basedir, cmd=['da_vs_turns', '-turnstep', str(da_turnstep), '-outfile'],
        #           python=python, ssh=ssh)
        # sixdb_cmd(jobname, basedir, cmd=['plot_da_vs_turns'], python=python, ssh=ssh)


class PostProcess(Stage):
    """
    Extracts the analysed data in the database and writes them to three tfs files:

    - All DA values
    - Statistics over angles, listed per seed (+ Seed 0 as over seeds and angles)
    - Statistics over seeds, listed per angle

    The statistics over the seeds are then plotted in a polar plot.
    All files are outputted to the ``sixjobs/autosix_output`` folder in the job directory.
    """
    def _run(self):
        post_process_da(self.jobname, self.basedir)


class Final(Stage):
    """ Just info about finishing this script and where to check the stagefile. """
    def _run(self):
        stage_file = get_stagefile_path(self.jobname, self.basedir)
        LOG.info(
            f"All stages run. Check stagefile {str(stage_file)} "
            "in case you want to rerun some stages."
        )
        raise StageSkip()



"""
HTCondor
--------

Constants for the HTCondor parameters.
"""
SHEBANG = "#!/bin/bash"
SUBFILE = "queuehtc.sub"
BASH_FILENAME = "Job"

HTCONDOR_JOBLIMIT = 100000

CMD_SUBMIT = "condor_submit"
JOBFLAVOURS = (
    "espresso",  # 20 min
    "microcentury",  # 1 h
    "longlunch",  # 2 h
    "workday",  # 8 h
    "tomorrow",  # 1 d
    "testmatch",  # 3 d
    "nextweek",  # 1 w
)

NOTIFICATIONS = ("always", "complete", "error", "never")


"""
SixDesk Submission Utils
---------------------------

Individual functions to call SixDesk functionality.
"""
import logging
from pathlib import Path
from typing import Union

from pylhc_submitter.constants.autosix import (
    MAD_TO_SIXTRACK_SH,
    RUNSIX_SH,
    RUNSTATUS_SH,
    SIXDB,
    get_sixjobs_path, StageSkip,
)
from pylhc_submitter.constants.external_paths import SIXDESK_UTILS
from pylhc_submitter.sixdesk_tools.utils import start_subprocess

LOG = logging.getLogger(__name__)


def submit_mask(jobname: str, basedir: Path, sixdesk: Path = SIXDESK_UTILS, ssh: str = None):
    """ Run the mask (probably Madx) and generate sixtrack input files.

    Args:
        jobname (str): Name of the Job
        basedir (Path): SixDesk Basefolder Location
        sixdesk (Path): Path to the SixDesk installation/repo
    """
    LOG.info("Submitting mask to run for sixtrack input generation.")
    sixjobs_path = get_sixjobs_path(jobname, basedir)
    start_subprocess([sixdesk / MAD_TO_SIXTRACK_SH, "-s"], cwd=sixjobs_path, ssh=ssh)
    LOG.info("Submitted mask-jobs.")


def check_sixtrack_input(jobname: str, basedir: Path, sixdesk: Path = SIXDESK_UTILS, ssh: str = None, resubmit: bool = False):
    """ Checks the generated input files needed by sixtrack and resubmits, if requested. """
    LOG.info("Checking if input files are present.")
    sixjobs_path = get_sixjobs_path(jobname, basedir)
    try:
        start_subprocess([sixdesk /MAD_TO_SIXTRACK_SH, "-c"], cwd=sixjobs_path, ssh=ssh)
    except OSError as e:
        if resubmit:
            LOG.info("Resubmitting mask to run wrong seeds for sixtrack input generation.")
            start_subprocess([sixdesk / MAD_TO_SIXTRACK_SH, "-w"], cwd=sixjobs_path, ssh=ssh)
            raise StageSkip("Resubmitted input generation jobs "
                            "(Not really an error, but the run is now interrupted).")
        else:
            raise StageSkip(
                "Checking input files failed. Check (debug-) logs. "
                "Maybe restart with 'resubmit' flag."
            ) from e
    else:
        LOG.info("Check for input files was successful.")


def submit_sixtrack(jobname: str, basedir: Path, python: Path = None,
                    sixdesk: Path = SIXDESK_UTILS, ssh: str = None, resubmit: bool = False):
    """ Generate simulation files and check if runnable and submit. """
    re_str = "Re-" if resubmit else ""
    LOG.info(f"{re_str}Submitting to sixtrack.")
    sixjobs_path = get_sixjobs_path(jobname, basedir)
    args = ["-i"] if resubmit else ["-a"]
    if python is not None:
        if not python.is_dir():
            python = python.parent
        args += ["-P", str(python)]
    try:
        start_subprocess([sixdesk / RUNSIX_SH] + args, cwd=sixjobs_path, ssh=ssh, check_log='exit status: 1')
    except OSError as e:
        raise StageSkip(
            f"{re_str}Submit to sixtrack for {jobname} ended in error."
            f" Input generation possibly not finished. Check your Scheduler."
        ) from e
    else:
        LOG.info(f"{re_str}Submitted jobs to Sixtrack")


def check_sixtrack_output(jobname: str, basedir: Path, python: Union[Path, str],
                          sixdesk: Path = SIXDESK_UTILS, ssh: str = None, resubmit: bool = False):
    """ Checks if the sixtrack output is all there. """
    LOG.info("Checking if sixtrack has finished.")
    sixjobs_path = get_sixjobs_path(jobname, basedir)
    try:
        start_subprocess([sixdesk / RUNSTATUS_SH], cwd=sixjobs_path, ssh=ssh)
    except OSError as e:
        if resubmit:
            submit_sixtrack(jobname, basedir, python=python, ssh=ssh, resubmit=True)
            raise StageSkip(
                f"Sixtrack for {jobname} seems to be incomplete."
                " Resubmitted incomplete sixtrack jobs."
                " Wait until they have finished and run again."
            )
        else:
            raise StageSkip(
                f"Sixtrack for {jobname} seems to be incomplete."
                f" Run possibly not finished. Check (debug-) log or your Scheduler."
            ) from e
    else:
        LOG.info("Sixtrack results are all present.")


def sixdb_load(jobname: str, basedir: Path, python: Union[Path, str],
               sixdesk: Path = SIXDESK_UTILS, ssh: str = None):
    """ Creates sixdb database and loads the study results into it. """
    LOG.info("Loading study into database.")
    sixjobs_path = get_sixjobs_path(jobname, basedir)
    try:
        start_subprocess([python, sixdesk / SIXDB, ".", "load_dir"], cwd=sixjobs_path, ssh=ssh)
    except OSError as e:
        raise StageSkip(f"Sixdb loading for {jobname} failed. Check (debug-) log.") from e
    else:
        LOG.info("Created database for study.")


def sixdb_cmd(jobname: str, basedir: Path, python: Union[Path, str], cmd: list,
              sixdesk: Path = SIXDESK_UTILS, ssh: str = None):
    """ Performs analysis on the sixdb database. """
    cmd_str = " ".join(cmd)
    LOG.info(f"Performing sixdb command `{cmd_str}`.")
    sixjobs_path = get_sixjobs_path(jobname, basedir)
    try:
        start_subprocess([python, sixdesk / SIXDB, jobname] + cmd, cwd=sixjobs_path, ssh=ssh)
    except OSError as e:
        raise StageSkip(f"SixBD command {cmd_str} for {jobname} failed. Check (debug-) log.") from e
    else:
        LOG.info(f"SixDB command '{cmd_str}' successfully run.")



"""
Extract Data From DataBase
-----------------------------

These functions operate on the SixDesk database and help to extract data
form it.

TODO: Implement extraction of data into ``.csv`` (and/or tfs?)
like fvanderv does.
"""
import logging
import sqlite3 as sql
from pathlib import Path
from typing import Any, Tuple, Iterable
from contextlib import contextmanager

import numpy as np
import pandas as pd
from generic_parser import DotDict
from matplotlib import pyplot as plt
from matplotlib import rcParams, lines as mlines
from scipy.interpolate import interp1d
from tfs import TfsDataFrame, write_tfs

from pylhc_submitter.constants.autosix import (
    get_database_path,
    get_tfs_da_path,
    get_tfs_da_seed_stats_path,
    get_tfs_da_angle_stats_path,
    get_autosix_results_path,
    HEADER_NTOTAL,
    HEADER_INFO,
    HEADER_HINT,
    MEAN,
    STD,
    MIN,
    MAX,
    N,
    SEED,
    ANGLE,
    ALOST1,
    ALOST2,
    AMP,
)

LOG = logging.getLogger(__name__)


def extract_da_data(jobname: str, basedir: Path) -> TfsDataFrame:
    """ Extract DA data directly from the database.

    Args:
        jobname (str): Name of the Job
        basedir (Path): SixDesk Basefolder Location
    """
    with _get_database(jobname, basedir) as db:
        df_da = pd.read_sql(
            "SELECT seed, angle, alost1, alost2, Amin, Amax FROM da_post ORDER BY seed, angle", db
        )

    df_da = df_da.rename(
        columns={
            "seed": SEED,
            "angle": ANGLE,
            "alost1": ALOST1,
            "alost2": ALOST2,
            "Amin": f"{MIN}{AMP}",
            "Amax": f"{MAX}{AMP}",
        }
    )
    return TfsDataFrame(df_da)


def extract_meta_data(jobname: str, basedir: Path) -> TfsDataFrame:
    """ Extract the meta-data directly from the database.

    Args:
        jobname (str): Name of the Job
        basedir (Path): SixDesk Basefolder Location
    """
    with _get_database(jobname, basedir) as db:
        df_da = pd.read_sql(
            "SELECT keyname, value FROM env", db
        )
    return TfsDataFrame(df_da)


@contextmanager
def _get_database(jobname, basedir):
    """Context to connect to DataBase and always close the connection afterwards.

    Args:
        jobname (str): Name of the Job
        basedir (Path): SixDesk Basefolder Location
    """
    db_path = get_database_path(jobname, basedir)
    db = sql.connect(db_path)
    yield db
    db.close()



"""
IO-Tools
--------

Tools for input and output.
"""
from datetime import datetime
from pathlib import Path
from typing import Iterable

from generic_parser.entry_datatypes import get_instance_faker_meta
from generic_parser.entrypoint_parser import save_options_to_config

from pylhc_submitter.constants.general import TIME

# Output -----------------------------------------------------------------------


def save_config(output_dir: Path, opt: dict, script: str):
    """
    Quick wrapper for ``save_options_to_config``.

    Args:
        output_dir (Path): Path to the output directory (does not need to exist).
        opt (dict): opt-structure to be saved.
        script (str): path/name of the invoking script (becomes name of the .ini)
                      usually ``__file__``.
    """
    output_dir.mkdir(parents=True, exist_ok=True)
    opt = convert_paths_in_dict_to_strings(opt)
    opt = escape_percentage_signs(opt)
    time = datetime.utcnow().strftime(TIME)
    save_options_to_config(output_dir / f"{script:s}_{time:s}.ini",
                           dict(sorted(opt.items()))
                           )


def escape_percentage_signs(dict_: dict) -> dict:
    """Escape all percentage signs.
    They are used for interpolation in the config-parser."""
    for key, value in dict_.items():
        if isinstance(value, dict):
            dict_[key] = escape_percentage_signs(value)
        elif isinstance(value, str):
            dict_[key] = value.replace("%(", "%%(")
    return dict_


def convert_paths_in_dict_to_strings(dict_: dict) -> dict:
    """Converts all Paths in the dict to strings, including those in iterables."""
    dict_ = dict_.copy()
    for key, value in dict_.items():
        if isinstance(value, Path):
            dict_[key] = str(value)
        else:
            try:
                list_ = list(value)
            except TypeError:
                pass
            else:
                has_changed = False
                for idx, item in enumerate(list_):
                    if isinstance(item, Path):
                        list_[idx] = str(item)
                        has_changed = True
                if has_changed:
                    dict_[key] = list_
    return dict_


# Input ------------------------------------------------------------------------


class PathOrStr(metaclass=get_instance_faker_meta(Path, str)):
    """A class that behaves like a Path when possible, otherwise like a string."""
    def __new__(cls, value):
        if value is None:
            return None

        if isinstance(value, str):
            value = value.strip("\'\"")  # behavior like dict-parser, IMPORTANT FOR EVERY STRING-FAKER
        return Path(value)


def make_replace_entries_iterable(replace_dict: dict) -> dict:
    """ Makes all entries in replace-dict iterable. """
    for key, value in replace_dict.items():
        if isinstance(value, str) or not isinstance(value, Iterable):
            replace_dict[key] = [value]
    return replace_dict


def keys_to_path(dict_, *keys):
    """ Convert all keys to Path, if they are not None. """
    for key in keys:
        value = dict_[key]
        dict_[key] = None if value is None else Path(value)
    return dict_



"""
Create SixDesk Workspace
-----------------------------------

Tools to setup the workspace for sixdesk.
"""
import re
import shutil
from dataclasses import asdict
from pathlib import Path
from typing import Union

import numpy as np
import logging

from generic_parser import DotDict

from pylhc_submitter.constants.autosix import (
    SETENV_SH,
    SIXENV_REQUIRED,
    SEED_KEYS,
    get_workspace_path,
    get_scratch_path,
    get_sixjobs_path,
    get_masks_path,
    get_mad6t_mask_path,
    get_mad6t1_mask_path,
    get_autosix_results_path,
    get_sysenv_path,
    get_sixdeskenv_path,
    SixDeskEnvironment, SIXENV_OPTIONAL,
)
from pylhc_submitter.constants.external_paths import SIXDESK_UTILS
from pylhc_submitter.sixdesk_tools.utils import start_subprocess

SYSENV_MASK = Path(__file__).parent / "mask_sysenv"
SIXDESKENV_MASK = Path(__file__).parent / "mask_sixdeskenv"

LOG = logging.getLogger(__name__)


# Main -------------------------------------------------------------------------


def create_job(jobname: str, basedir: Path, executable: Union[Path, str], mask_text: str,
               sixdesk: Path = SIXDESK_UTILS, ssh: str = None, **kwargs):
    """ Create environment and individual jobs/masks for SixDesk to send to HTC.

    Keyword Args:
        Need to contain all replacements for sixdeskenv and the mask.
    """
    _create_workspace(jobname, basedir, sixdesk=sixdesk, ssh=ssh)
    _create_sysenv(jobname, basedir, binary_path=executable)
    _create_sixdeskenv(jobname, basedir, **kwargs)
    _write_mask(jobname, basedir, mask_text, **kwargs)
    LOG.info("Workspace prepared.")


def init_workspace(jobname: str, basedir: Path, sixdesk: Path = SIXDESK_UTILS, ssh: str = None):
    """ Initializes the workspace with sixdeskenv and sysenv. """
    sixjobs_path = get_sixjobs_path(jobname, basedir)
    start_subprocess([sixdesk / SETENV_SH, "-s"], cwd=sixjobs_path, ssh=ssh)
    LOG.info("Workspace initialized.")


def remove_twiss_fail_check(jobname: str, basedir: Path):
    """ Comments out the "Twiss fail" check from mad6t.sh """
    LOG.info("Applying twiss-fail hack.")
    for mad6t_path in (
        get_mad6t_mask_path(jobname, basedir),
        get_mad6t1_mask_path(jobname, basedir),
    ):
        with open(mad6t_path, "r") as f:
            lines = f.readlines()

        check_started = False
        for idx, line in enumerate(lines):
            if line.startswith('grep -i "TWISS fail"'):
                check_started = True

            if check_started:
                lines[idx] = f"# {line}"
                if line.startswith("fi"):
                    break
        else:
            LOG.info(f"'TWISS fail' not found in {mad6t_path.name}")
            continue

        with open(mad6t_path, "w") as f:
            f.writelines(lines)


def fix_pythonfile_call(jobname: str, basedir: Path):
    """ Removes '<' in the `binary file` line in mad6t.sh so __file__ works. """
    LOG.info("Applying python-file call fix.")
    for mad6t_path in (
            get_mad6t_mask_path(jobname, basedir),
            get_mad6t1_mask_path(jobname, basedir),
    ):
        with open(mad6t_path, "r") as f:
            lines = f.readlines()

        for idx, line in enumerate(lines):
            if line.startswith('$MADX_PATH/$MADX'):
                lines[idx] = f'$MADX_PATH/$MADX $junktmp/$filejob."$i" > $filejob.out."$i"\n'
                break
        else:
            raise IOError(f"'$MADX_PATH/$MADX' line not found in {mad6t_path.name}")

        with open(mad6t_path, "w") as f:
            f.writelines(lines)


def set_max_materialize(sixdesk: Path, max_materialize: int = None):
    """ Adds the ``max_materialize`` option into the htcondor sixtrack
    submission-file."""
    if max_materialize is None:
        return

    LOG.info(f"Setting max_materialize for SixTrack to {max_materialize}.")
    sub_path = sixdesk / "utilities" / "templates" / "htcondor" / "htcondor_run_six.sub"
    sub_content = sub_path.read_text()

    # Remove whole max_materialize line if present
    if max_materialize == 0:
        if "max_materialize" in sub_content:
            LOG.info("'max_materialize' already set. Removing.")
            sub_content = re.sub(r"max_materialize\s*=\s*\d+\s*", "", sub_content)
        else:
            LOG.debug("'max_materialize' is already not present (as desired).")

    # Set/replace with number
    else:
        max_materialize_str = f"max_materialize = {max_materialize:d}"
        if "max_materialize" in sub_content:
            LOG.info("max_materialize already set. Replacing it with new number.")
            sub_content = re.sub(r"max_materialize\s*=\s*\d+", max_materialize_str, sub_content)
        else:
            sub_content = sub_content.replace("\nqueue", f"\n{max_materialize_str}\nqueue")

    # Write out
    try:
        sub_path.write_text(sub_content)
    except IOError as e:
        raise IOError(f"Could not write to {sub_path!s}. `max_materialization` could not be set.\n"
                      f"Remove option or use a SixDesk with writing rights.") from e


# Helper -----------------------------------------------------------------------


def _create_workspace(jobname: str,  basedir: Path, sixdesk: Path = SIXDESK_UTILS, ssh: str = None):
    """ Create workspace structure (with default files). """
    workspace_path = get_workspace_path(jobname, basedir)
    scratch_path = get_scratch_path(basedir)
    LOG.info(f'Creating new workspace in "{str(workspace_path)}"')

    if workspace_path.exists():
        LOG.warning(f'Workspace in "{str(workspace_path)}" already exists. ')
        LOG.info("Do you want to delete the old workspace? [y/N]")
        user_answer = input()
        if user_answer.lower().startswith("y"):
            shutil.rmtree(workspace_path)
            try:
                shutil.rmtree(scratch_path)
            except FileNotFoundError:
                pass
        else:
            LOG.warning("Keeping Workspace as-is.")
            return

    scratch_path.mkdir(parents=True, exist_ok=True)

    # create environment with all necessary files
    # _start_subprocess(['git', 'clone', GIT_REPO, basedir])
    start_subprocess([sixdesk / SETENV_SH, "-N", workspace_path.name], cwd=basedir, ssh=ssh)

    # create autosix results folder.
    # Needs to be done after above command (as it crashes if folder exists)
    # but before end of this stage (as it needs to write the stagefile)
    get_autosix_results_path(jobname, basedir).mkdir(exist_ok=True, parents=True)


def _create_sixdeskenv(jobname: str, basedir: Path, **kwargs):
    """ Fills sixdeskenv mask and copies it to workspace """
    workspace_path = get_workspace_path(jobname, basedir)
    scratch_path = get_scratch_path(basedir)
    sixdeskenv_path = get_sixdeskenv_path(jobname, basedir)

    missing = [key for key in SIXENV_REQUIRED if key not in kwargs.keys()]
    if len(missing):
        raise ValueError(f"The following keys are required but missing {missing}.")

    sixenv_variables = SixDeskEnvironment(
        JOBNAME=jobname,
        WORKSPACE=workspace_path.name,
        BASEDIR=str(basedir),
        SCRATCHDIR=str(scratch_path),
        TURNSPOWER=np.log10(kwargs["TURNS"]),
        **{k: v for k, v in kwargs.items() if k in SIXENV_REQUIRED + SIXENV_OPTIONAL},
    )

    sixenv_text = SIXDESKENV_MASK.read_text()
    sixdeskenv_path.write_text(sixenv_text % asdict(sixenv_variables))
    LOG.debug("sixdeskenv written.")


def _create_sysenv(jobname: str, basedir: Path, binary_path: Path):
    """ Fills sysenv mask and copies it to workspace """
    LOG.info(f"Chosen binary for mask '{str(binary_path)}'")
    sysenv_path = get_sysenv_path(jobname, basedir)
    sysenv_replace = dict(
        MADXPATH=str(binary_path.parent),
        MADXBIN=binary_path.name,
    )
    sysenv_text = SYSENV_MASK.read_text()
    sysenv_path.write_text(sysenv_text % sysenv_replace)
    LOG.debug("sysenv written.")


def _write_mask(jobname: str, basedir: Path, mask_text: str, **kwargs):
    """ Fills mask with arguments and writes it out. """
    masks_path = get_masks_path(jobname, basedir)
    seed_range = [kwargs.get(key, getattr(SixDeskEnvironment, key)) for key in SEED_KEYS]

    if seed_range.count(None) == 1:
        raise ValueError(
            "First- or Lastseed is set, but the other one is deactivated. " "Set or unset both."
        )

    if ("%SEEDRAN" not in mask_text) and ("%SEEDRAN" not in kwargs.values()) and any(seed_range):
        raise ValueError(
            "First- and Lastseed are set, but no seed-variable '%SEEDRAN' found in mask."
        )

    mask_text = mask_text.replace("%SEEDRAN", "#!#SEEDRAN")  # otherwise next line will complain
    mask_filled = mask_text % kwargs
    mask_filled = mask_filled.replace(
        "#!#SEEDRAN", "%SEEDRAN"
    )  # bring seedran back for sixdesk seed-loop

    with open(masks_path / f"{jobname}.mask", "w") as mask_out:
        mask_out.write(mask_filled)



"""
Autosix
-------

Collections of constants and paths used in autosix.

:module: constants.autosix
:author: jdilly

"""
from dataclasses import dataclass, fields, MISSING
from pathlib import Path
from typing import Union

import logging
from pylhc_submitter.constants.external_paths import MADX_BIN, SIXDESK_UTILS
from pylhc_submitter.constants.general import PMASS

LOG = logging.getLogger(__name__)

# Program Paths ----------------------------------------------------------------
UTILITIES_DIR: Path = Path("utilities")
BASH_DIR: Path = UTILITIES_DIR / "bash"
SETENV_SH: Path = BASH_DIR / "set_env.sh"
MAD_TO_SIXTRACK_SH: Path = BASH_DIR / "mad6t.sh"
RUNSIX_SH: Path = BASH_DIR / "run_six.sh"
RUNSTATUS_SH: Path = BASH_DIR / "run_status"
DOT_PROFILE: Path = BASH_DIR / "dot_profile"
SIXDB: Path = UTILITIES_DIR / "externals" / "SixDeskDB" / "sixdb"
SIXDESKLOCKFILE: Path = "sixdesklock"

# Constants and Requirements ---------------------------------------------------

HEADER_BASEDIR = "BASEDIR"


# AutoSix Environment (also defines defaults) ---
@dataclass
class AutoSixEnvironment:
    mask_text: str
    working_directory: Path
    executable: Union[str, Path] = MADX_BIN
    python2: Union[str, Path] = None
    python3: Union[str, Path] = "python3"
    da_turnstep: int = 100
    sixdesk_directory: Path = SIXDESK_UTILS
    unlock: bool = False
    max_stage: 'Stage' = None
    ssh: str = None
    stop_workspace_init: bool = False
    apply_mad6t_hacks: bool = False
    resubmit: bool = False
    max_materialize: int = None


# Sixenv ---
class _SetDynamicallyType:
    pass
SetDynamically = _SetDynamicallyType()
"""Sentinel value defining a parameter that is set later dynamically."""

@dataclass
class SixDeskEnvironment:
    TURNS: int
    AMPMIN: int
    AMPMAX: int
    AMPSTEP: int
    ANGLES: int
    JOBNAME: str = SetDynamically
    WORKSPACE: str = SetDynamically
    BASEDIR: str = SetDynamically
    SCRATCHDIR: str = SetDynamically
    TURNSPOWER: int = SetDynamically
    RESUBMISSION: int = 0  # 0: never, 1: if fort.10, 2: always
    PLATFORM: str = "HTCondor"
    LOGLEVEL: int = 0  # 0: basic + errors , 1: + info, >=2: + debug
    FIRSTSEED: int = 1
    LASTSEED: int = 60
    RUNTYPE: str = "col"  # 'col' or 'inj'
    NPAIRS: int = 30  # 1-32 particle pairs
    EMITTANCE: float = 3.75  # normalized emittance
    DIMENSIONS: int = 6  # Phase-Space dimensions
    WRITEBINS: int = 500
    ENERGY: float = None
    GAMMA: float = SetDynamically  # calculated from ENERGY, needs to be here so asdict() converts it

    def __post_init__(self):
        # Check Runtype, Energy and Gamma ---
        energy_map = {'inj': 450_000., 'col': 7000_000.}
        if self.RUNTYPE not in energy_map.keys():
            raise ValueError(f"RUNTYPE needs to be one of : {list(energy_map.keys())}")

        if self.ENERGY is None:
            energy = energy_map[self.RUNTYPE]
            LOG.debug(f"Energy for '{self.RUNTYPE}' defaults to {energy}")
            self.ENERGY = energy

        if self.ENERGY >= 6500_000 and self.RUNTYPE != list(energy_map.keys())[1]:
            LOG.warning(f"Runtype is {self.RUNTYPE}, yet energy is set to {self.ENERGY}. Are you sure?")

        self.GAMMA = self.ENERGY / 1000 / PMASS

        # Check Seeds ---
        if any(getattr(self, key) is None for key in SEED_KEYS):  # set by user to None
            for key in SEED_KEYS:
                setattr(self, key, 0)

        # the following checks are limits of SixDesk in 2020
        # and might be fixed upstream in the future
        if self.AMPMAX < self.AMPMIN:
            raise ValueError("Given AMPMAX is smaller than AMPMIN.")

        if (self.AMPMAX - self.AMPMIN) % self.AMPSTEP:
            raise ValueError("The amplitude range need to be dividable by the amplitude steps!")

        if not self.ANGLES % 2:
            raise ValueError("The number of angles needs to be an uneven one.")


SIXENV_REQUIRED = [f.name for f in fields(SixDeskEnvironment) if (f.default is MISSING)]  # required by user
SIXENV_OPTIONAL = [f.name for f in fields(SixDeskEnvironment) if (f.default is not MISSING) and (f.default is not SetDynamically)]
SEED_KEYS = ["FIRSTSEED", "LASTSEED"]

# SixDB and Postprocess ---

SIXTRACK_OUTPUT_FILES = "fort.10.gz", "Sixout.zip"
SIXTRACK_INPUT_CHECK_FILES = "JOB_NOT_YET_STARTED", "JOB_NOT_YET_COMPLETED"

HEADER_NTOTAL, HEADER_INFO, HEADER_HINT = "NTOTAL", "INFO", "HINT"
MEAN, STD, MIN, MAX, N = "MEAN", "STD", "MIN", "MAX", "N"
SEED, ANGLE, ALOST1, ALOST2, AMP = "SEED", "ANGLE", "ALOST1", "ALOST2", "A"


# Errors ---


class StageSkip(Exception):
    """ Indicates that the stage was not completed or skipped entirely.
    This can be due to an error or on purpose (e.g. user interaction before
    restart). """
    pass


class StageStop(Exception):
    """ A signal sent at the end of a Stage indicating, that it has succeeded
    and that any iteration should be stopped after this Stage as the jobs have
    been submitted and the user needs to wait for them to finish. """
    pass


# Workspace Paths --------------------------------------------------------------

# Input ---
def get_workspace_path(jobname: str, basedir: Path) -> Path:
    return basedir / f"workspace-{jobname}"


def get_scratch_path(basedir: Path) -> Path:
    return basedir / f"scratch-0"


def get_sixjobs_path(jobname: str, basedir: Path) -> Path:
    return get_workspace_path(jobname, basedir) / "sixjobs"


def get_sixdeskenv_path(jobname: str, basedir: Path) -> Path:
    return get_sixjobs_path(jobname, basedir) / "sixdeskenv"


def get_sysenv_path(jobname: str, basedir: Path) -> Path:
    return get_sixjobs_path(jobname, basedir) / "sysenv"


def get_masks_path(jobname: str, basedir: Path) -> Path:
    return get_sixjobs_path(jobname, basedir) / "mask"


def get_track_path(jobname: str, basedir: Path) -> Path:
    return get_sixjobs_path(jobname, basedir) / "track"


def get_database_path(jobname: str, basedir: Path) -> Path:
    return get_sixjobs_path(jobname, basedir) / f"{jobname}.db"


def get_sixtrack_input_path(jobname: str, basedir: Path) -> Path:
    return get_sixjobs_path(jobname, basedir) / "sixtrack_input"


def get_mad6t_mask_path(jobname: str, basedir: Path) -> Path:
    return get_sixtrack_input_path(jobname, basedir) / "mad6t.sh"


def get_mad6t1_mask_path(jobname: str, basedir: Path) -> Path:
    return get_sixtrack_input_path(jobname, basedir) / "mad6t1.sh"


# Output ---


def get_autosix_results_path(jobname: str, basedir: Path) -> Path:
    return get_sixjobs_path(jobname, basedir) / "autosix_output"


def get_stagefile_path(jobname: str, basedir: Path) -> Path:
    return get_autosix_results_path(jobname, basedir) / "stages_completed.txt"


def get_tfs_da_path(jobname: str, basedir: Path) -> Path:
    return get_autosix_results_path(jobname, basedir) / f"{jobname}_da.tfs"


def get_tfs_da_seed_stats_path(jobname: str, basedir: Path) -> Path:
    return get_autosix_results_path(jobname, basedir) / f"{jobname}_da_per_seed.tfs"


def get_tfs_da_angle_stats_path(jobname: str, basedir: Path) -> Path:
    return get_autosix_results_path(jobname, basedir) / f"{jobname}_da_per_angle.tfs"




""" 
Job Submitter IO-Tools
----------------------

Tools for input and output for the job-submitter.
"""
import itertools
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Sequence, Tuple, Union

import numpy as np
import pandas as pd
import tfs

from pylhc_submitter.constants.htcondor import HTCONDOR_JOBLIMIT
from pylhc_submitter.constants.job_submitter import (COLUMN_DEST_DIRECTORY, COLUMN_JOB_DIRECTORY,
                                                     COLUMN_JOBID, JOBDIRECTORY_PREFIX,
                                                     JOBSUMMARY_FILE, SCRIPT_EXTENSIONS)
from pylhc_submitter.submitter import htc_utils
from pylhc_submitter.submitter.mask import (create_job_scripts_from_mask, generate_jobdf_index,
                                            is_mask_file)

LOG = logging.getLogger(__name__)


JobNamesType = Sequence[Union[str, int]]


@dataclass
class CreationOpts:
    """ Options for creating jobs. """
    working_directory: Path         # Path to working directory (afs)
    mask: Union[Path, str]          # Path to mask file or mask-string
    jobid_mask: str                 # Mask for jobid
    replace_dict: Dict[str, Any]    # Replace-dict
    output_dir: Path                # Path to local output directory
    output_destination: Union[Path, str]  # Path or URI to remote output directory (e.g. eos)
    append_jobs: bool               # Append jobs to existing jobs
    resume_jobs: bool               # Resume jobs that have already run/failed/got interrupted
    executable: str                 # Name of executable to call the script (from mask)
    check_files: Sequence[str]      # List of output files to check for success
    script_arguments: Dict[str, Any]  # Arguments to pass to script
    script_extension: str           # Extension of the script to run

    def should_drop_jobs(self) -> bool:
        """ Check if jobs should be dropped after creating the whole parameter space, 
        e.g. because they already exist. """
        return self.append_jobs or self.resume_jobs



def create_jobs(opt: CreationOpts) -> tfs.TfsDataFrame:
    """Main function to prepare all the jobs and folder structure.
    This greates the value-grid based on the replace-dict and
    checks for existing jobs (if so desired).
    A job-dataframe is created - and written out - containing all the information and
    its values are used to generate the job-scripts.
    It also creates bash-scripts to call the executable for the job-scripts. 

    Args:
        opt (CreationOpts): Options for creating jobs 

    Returns:
        tfs.TfsDataFrame: The job-dataframe containing information for all jobs. 
    """
    LOG.debug("Creating Jobs.")

    # Generate product of replace-dict and compare to existing jobs  ---
    parameters, values_grid, prev_job_df = _generate_parameter_space(
        replace_dict=opt.replace_dict,
        append_jobs=opt.append_jobs,
        cwd=opt.working_directory,
    )

    # Check new jobs ---
    njobs = len(values_grid)
    if njobs == 0:
        raise ValueError(f"No (new) jobs found!")

    if njobs > HTCONDOR_JOBLIMIT:
        LOG.warning(
            f"You are attempting to submit an important number of jobs ({njobs})."
            "This can be a high stress on your system, make sure you know what you are doing."
        )

    LOG.debug(f"Initial number of jobs: {njobs:d}")

    # Generate new job-dataframe ---
    job_df = tfs.TfsDataFrame(
        index=generate_jobdf_index(prev_job_df, opt.jobid_mask, parameters, values_grid),
        columns=parameters,
        data=values_grid,
    )
    job_df = tfs.concat([prev_job_df, job_df], sort=False, how_headers='left')

    # Setup folders ---
    job_df = create_folders(job_df, opt.working_directory, opt.output_destination)

    # Create scripts ---
    if is_mask_file(opt.mask):
        LOG.debug("Creating all jobs from mask.")
        script_extension = _get_script_extension(opt.script_extension, opt.executable, opt.mask)
        job_df = create_job_scripts_from_mask(
            job_df, opt.mask, parameters, script_extension
        )

    LOG.debug("Creating shell scripts.")
    job_df = htc_utils.write_bash(
        job_df,
        output_dir=opt.output_dir,
        executable=opt.executable,
        cmdline_arguments=opt.script_arguments,
        mask=opt.mask,
    )

    # Convert paths to strings and write df to file ---
    job_df[COLUMN_JOB_DIRECTORY] = job_df[COLUMN_JOB_DIRECTORY].apply(str)
    if COLUMN_DEST_DIRECTORY in job_df.columns:
        job_df[COLUMN_DEST_DIRECTORY] = job_df[COLUMN_DEST_DIRECTORY].apply(str)

    tfs.write(str(opt.working_directory / JOBSUMMARY_FILE), job_df, save_index=COLUMN_JOBID)
    
    # Drop already run jobs ---
    dropped_jobs = []
    if opt.should_drop_jobs():
        job_df, dropped_jobs = _drop_already_run_jobs(
            job_df, opt.output_dir, opt.check_files
        )
    return job_df, dropped_jobs


def create_folders(job_df: tfs.TfsDataFrame, working_directory: Path, 
                   destination_directory: Union[Path, str] = None) -> tfs.TfsDataFrame:
    """Create the folder-structure in the given working directory and the 
    destination directory if given.
    This creates a folder per job in which then the job-scripts and bash-scripts
    can be stored later.

    Args:
        job_df (tfs.TfsDataFrame): DataFrame containing all the job-information
        working_directory (Path): Path to the working directory
        destination_directory (Path, optional): Path to the destination directory, 
        i.e. the directory to copy the outputs to manually. Defaults to None.

    Returns:
        tfs.TfsDataFrame: The job-dataframe again, but with the added paths to the job-dirs.
    """
    LOG.debug("Setting up folders: ")
    
    jobname = f"{JOBDIRECTORY_PREFIX}.{{0}}"
    job_df[COLUMN_JOB_DIRECTORY] = [working_directory / jobname.format(id_) for id_ in job_df.index]

    for job_dir in job_df[COLUMN_JOB_DIRECTORY]:
        job_dir.mkdir(exist_ok=True)
        LOG.debug(f"   created '{job_dir}'.")

    if destination_directory:
        dest_path = uri_to_path(destination_directory)
        dest_path.mkdir(parents=True, exist_ok=True)

        server = get_server_from_uri(destination_directory)
        job_df[COLUMN_DEST_DIRECTORY] = [f"{server}{dest_path / jobname.format(id_)}" for id_ in job_df.index]

        # Make some symlinks for easy navigation---
        # Output directory -> Working Directory
        sym_submission = dest_path / Path('SUBMISSION_DIR')
        sym_submission.unlink(missing_ok=True)
        sym_submission.symlink_to(working_directory.resolve(), target_is_directory=True)

        # Working Directory -> Output Directory
        sym_destination = working_directory / Path('OUTPUT_DIR')
        sym_destination.unlink(missing_ok=True)
        sym_destination.symlink_to(dest_path.resolve(), target_is_directory=True)

        # Create output dirs per job ---
        for job_dest_dir in job_df[COLUMN_DEST_DIRECTORY]:
            uri_to_path(job_dest_dir).mkdir(exist_ok=True)
            LOG.debug(f"   created '{job_dest_dir}'.")

    return job_df


def is_eos_uri(path: Union[Path, str, None]) -> bool:
    """ Check if the given path is an EOS-URI as `eos cp` only works with those.
    E.g.: root://eosuser.cern.ch//eos/user/a/anabramo/banana.txt
    
    This function does not check the double slashes,
    to avoid having the user pass a malformed path by accident and then 
    assuming it is just a path. This is tested for in 
    :func:`pylhc_submitter.job_submitter.check_opts`.
    """
    if path is None:
        return False

    parts = Path(path).parts 
    return (
        len(parts) >= 3  # at least root:, server, path
        and
        parts[0].endswith(':')
        and
        parts[2] == 'eos'
    )


def uri_to_path(path: Union[Path, str]) -> Path:
    """ Strip EOS path information from a path.
    EOS paths for HTCondor can be given as URI. Strip for direct writing.
    E.g.: root://eosuser.cern.ch//eos/user/a/anabramo/banana.txt
    """
    path = Path(path)
    parts = path.parts
    if parts[0].endswith(':'):
        # the first two parts are host info, e.g `file: //host/path`
        return Path('/', *parts[2:])
    return path 


def get_server_from_uri(path: Union[Path, str]) -> str:
    """ Get server information from a path.
    E.g.: root://eosuser.cern.ch//eos/user/a/ -> root://eosuser.cern.ch/
    """
    path_part = uri_to_path(path)
    if path_part == Path(path):
        return ""
    
    server_part = str(path).replace(str(path_part), '')
    if server_part.endswith("//"):
        server_part = server_part[:-1]
    return server_part


def print_stats(new_jobs: JobNamesType, finished_jobs: JobNamesType):
    """Print some quick statistics."""
    text = [
        "\n------------- QUICK STATS ----------------"
        f"Jobs total:{len(new_jobs) + len(finished_jobs):d}",
        f"Jobs to run: {len(new_jobs):d}",
        f"Jobs already finished: {len(finished_jobs):d}",
        "---------- JOBS TO RUN: NAMES -------------"
    ]
    text += [str(job_name) for job_name in new_jobs]
    text += ["--------- JOBS FINISHED: NAMES ------------"]
    text += [str(job_name) for job_name in finished_jobs]
    LOG.info("\n".join(text))


def _generate_parameter_space(
        replace_dict: Dict[str, Any], append_jobs: bool, cwd: Path
    ) -> Tuple[List[str], np.ndarray, tfs.TfsDataFrame]:
    """ Generate parameter space from replace-dict, check for existing jobs. """
    LOG.debug("Generating parameter space from replace-dict.")
    parameters = list(replace_dict.keys())
    values_grid = _generate_values_grid(replace_dict)
    if not append_jobs:
        return parameters, values_grid, tfs.TfsDataFrame()

    jobfile_path = cwd / JOBSUMMARY_FILE
    try:
        prev_job_df = tfs.read(str(jobfile_path.absolute()), index=COLUMN_JOBID)
    except FileNotFoundError as filerror:
        raise FileNotFoundError(
            "Cannot append jobs, as no previous jobfile was found at " f"'{jobfile_path}'"
        ) from filerror
    new_jobs_mask = [elem not in prev_job_df[parameters].values for elem in values_grid]
    values_grid = values_grid[new_jobs_mask]

    return parameters, values_grid, prev_job_df


def _generate_values_grid(replace_dict: Dict[str, Any]) -> np.ndarray:
    """ Creates an array of the inner-product of the replace-dict. """
    return np.array(list(itertools.product(*replace_dict.values())), dtype=object)


def _drop_already_run_jobs(
        job_df: tfs.TfsDataFrame, output_dir: str, check_files: str
    ) -> Tuple[tfs.TfsDataFrame, List[str]]:
    """ Check for jobs that have already been run and drop them from current job_df. """
    LOG.debug("Dropping already finished jobs.")
    finished_jobs = [
        idx
        for idx, row in job_df.iterrows()
        if _job_was_successful(row, output_dir, check_files)
    ]

    LOG.info(
        f"{len(finished_jobs):d} of {len(job_df.index):d}"
        " Jobs have already finished and will be skipped."
    )

    job_df = job_df.drop(index=finished_jobs)
    return job_df, finished_jobs


def _job_was_successful(job_row: pd.Series, output_dir: str, files: Sequence[str]) -> bool:
    """ Determines if the job was successful. 
    
    Args:
        job_row (pd.Series): row from the job_df
        output_dir (str): Name of the (local) output directory
        files (List[str]): list of files that should have been generated
    """
    job_dir = job_row.get(COLUMN_DEST_DIRECTORY) or job_row[COLUMN_JOB_DIRECTORY]
    output_dir = Path(job_dir, output_dir)
    success = output_dir.is_dir() and any(output_dir.iterdir())
    if success and files is not None and len(files):
        for f in files:
            success &= len(list(output_dir.glob(f))) > 0
    return success


def _get_script_extension(script_extension: str, executable: Path, mask: Path) -> str:
    """ Returns the extension of the script to run based on 
    either the given value, its executable or the mask. """
    if script_extension is not None:
        return script_extension
    return SCRIPT_EXTENSIONS.get(executable, mask.suffix)



"""
Mask Resolver
-------------

This module provides functionality to resolve and write script masks for ``HTCondor`` jobs
submission.
"""
import logging
import re
from pathlib import Path
from typing import Iterable, List, Sequence, Set, Union

import pandas as pd
from numpy.typing import ArrayLike

from pylhc_submitter.constants.job_submitter import COLUMN_JOB_DIRECTORY, COLUMN_JOB_FILE

LOG = logging.getLogger(__name__)


def create_job_scripts_from_mask(
    job_df: pd.DataFrame, maskfile: Path, replace_keys: dict, file_ext: str
) -> pd.DataFrame:
    """
    Takes path to mask file, list of parameter to be replaced and pandas dataframe containg per job
    the job directory where processed mask is to be put, and columns containing the parameter values
    with column named like replace parameters. Job directories have to be created beforehand.
    Processed (madx) mask has the same filename as mask but with the given file extension.
    Input Dataframe is returned with additional column containing path to the processed script
    files.

    Args:
        job_df (pd.DataFrame): Job parameters as defined in description.
        maskfile: `Path` object to the mask file.
        replace_keys: keys to be replaced (must correspond to columns in ``job_df``).
        file_ext: file extention to use (defaults to **madx**).

    Returns:
        The provided ``job_df`` but with added path to the scripts.
    """
    with maskfile.open("r") as mfile:
        template = mfile.read()

    jobname = maskfile.with_suffix("").name
    jobs = [None] * len(job_df)
    for idx, (jobid, values) in enumerate(job_df.iterrows()):
        jobfile_fullpath = (Path(values[COLUMN_JOB_DIRECTORY]) / jobname).with_suffix(file_ext)

        with jobfile_fullpath.open("w") as job_file:
            job_file.write(template % dict(zip(replace_keys, values[list(replace_keys)])))
        jobs[idx] = jobfile_fullpath.name
    job_df[COLUMN_JOB_FILE] = jobs
    return job_df


def find_named_variables_in_mask(mask: str) -> Set[str]:
    """ Find all variable-names in the mask. """
    return set(re.findall(r"%\((\w+)\)", mask))


def check_percentage_signs_in_mask(mask: str) -> None:
    """ Checks for '%' in the mask, that are not replacement variables. """
    cleaned_mask = re.sub(r"%\((\w+)\)", "", mask)
    n_signs = cleaned_mask.count("%")
    if n_signs == 0:
        return

    # Help the user find the %
    for idx, line in enumerate(cleaned_mask.split("\n")):
        if "%" in line:
            positions = [str(i) for i, char in enumerate(line) if char == "%"]
            LOG.error(f"Problematic '%' sign(s) in line {idx}, pos {' ,'.join(positions)}.")
    raise KeyError(f"{n_signs} problematic '%' signs found in template. Please remove.")


def generate_jobdf_index(old_df: pd.DataFrame, jobid_mask: str, keys: Sequence[str], values: ArrayLike
    ) -> Union[List[str], Iterable[int]]:
    """ Generates index for jobdf from mask for job_id naming. 
    
    Args:
        old_df (pd.DataFrame): Existing jobdf.
        jobid_mask (str): Mask for naming the jobs.
        keys (Sequence[str]): Keys to be replaced in the mask.
        values (np.array_like): Values-Grid to be replaced in the mask.
    
    Returns:
        List[str]: Index for jobdf, either list of strings (the filled jobid_masks) or integer-range.
    """
    if not jobid_mask:
        # Use integer-range as index, if no mask is given
        # Start with last index if old_df is not None.
        nold = len(old_df.index) if old_df is not None else 0
        start = nold-1 if nold > 0 else 0
        return range(start, start + values.shape[0])

    # Fill job-id mask
    return [jobid_mask % dict(zip(keys, v)) for v in values]


def is_mask_file(mask: str) -> bool:
    """ Check if given string points to a file. """
    try:
        return Path(mask).is_file()
    except OSError:
        return False


def is_mask_string(mask: str) -> bool:
    """ Checks that given string does not point to a file. """
    return not is_mask_file(mask)


if __name__ == "__main__":
    raise EnvironmentError(f"{__file__} is not supposed to run as main.")



"""
AutoSix
-------

``AutoSix`` is a wrapper to automatically perform the necessary setup and steps needed for ``SixDesk`` use.

The functionality is similar to the ``pylhc_submitter.job_submitter`` in that the inner product of a ``replace_dict``
is used to automatically create a set of job-directories to gather the data.
To avoid conflicts, each of these job-directories is a ``SixDesk`` workspace,
meaning there can be only one study per directory.
Beware that the ``max_materialize`` limit is set for each of these workspaces
individually, not for all Jobs together (i.e. it should be <=MAX_USER_JOBS / NUMBER_OF_WORKSPACES).

The ``replace_dict`` contains variables for your mask as well as variables for the SixDesk environment.
See the description of ``replace_dict`` below.

In any other way, these *special* variables behave like normal variables and can also be inserted in your mask.
They are also looped over in the same manner as any other variable (if given as a list).

For additional information and guides, see the `AutoSix page
<https://pylhc.github.io/packages/pylhcsubmitter/autosix/>`_
on the ``OMC`` documentation site.

.. admonition:: Problems with *search.f90*
    :class: warning

    If you run into

    .. code-block:: bash

       ImportError: cannot import name 'search' from partially initialized module 'sixdeskdb'

    that means that the :file:`SixDesk/utilities/externals/SixDeskDB/sixdeskdb/search.f90`
    has not been compiled for your current python and OS version (indicated by XXX below).
    There are two ways to do this:

    **a)** Run

    .. code-block:: bash

        python -m numpy.f2py -c search.f90 -f search

    with your desired python version
    (a shortcut ``f2py`` might be available if you run in an activated venv).
    Copy (or symlink) the resulting :file:`search.cpython-XXX.so` file into the
    :file:`sixdeskdb` folder
    (if it is not already there because you ran from that folder)

    **b)** Run

    .. code-block:: bash

        python setup.py build_ext

    on the :file:`setup.py` in :file:`SixDesk/utilities/externals/SixDeskDB`.
    Then copy (or symlink) the resulting :file:`build/lib.XXX/sixdeskdb/search.cpython-XXX.so` into
    the :file:`sixdeskdb` folder (the name needs to stay as it is).


Arguments:

*--Required--*

- **mask** *(Path)*:

    Path to the program's mask to use.


- **replace_dict** *(DictAsString)*:

    Dict with keys of the strings to be replaced in the mask (required) as
    well as the mask_sixdeskenv and mask_sysenv files in the sixdesk_tools
    module. Required fields are TURNS, AMPMIN, AMPMAX, AMPSTEP, ANGLES.
    Optional fields are RESUBMISSION, PLATFORM, LOGLEVEL, FIRSTSEED,
    LASTSEED, RUNTYPE, NPAIRS, EMITTANCE, DIMENSIONS, WRITEBINS, ENERGY.
    These keys can also be used in the mask if needed. The values of this
    dict are lists of values to replace these or single entries.


- **working_directory** *(Path)*:

    Directory where data should be put into.


*--Optional--*

- **apply_mad6t_hacks**:

    Apply two hacks: Removes '<' in binary call and ignore the check for
    'Twiss fail' in the submission file. This is hack needed in case this
    check greps the wrong lines, e.g. in madx-comments. USE WITH CARE!!

    action: ``store_true``


- **da_turnstep** *(int)*:

    Step between turns used in DA-vs-Turns plot.

    default: ``100``


- **executable** *(PathOrStr)*:

    Path to executable or 'madx', 'python2', 'python3' to use the OMC
    default paths on AFS.Defaults to the latest MADX-Binary on AFS.

    default: ``/afs/cern.ch/user/m/mad/bin/madx``


- **jobid_mask** *(str)*:

    Mask-String to name jobs, with placeholders from the ``replace_dict``
    keys.


- **max_materialize** *(int)*:

    Maximum jobs to be materialized in scheduler (per SixDesk Workspace!)..
    Here: ``None`` leaves the settings as defined in the SixDesk
    htcondor_run_six.sub template and ``0`` removes it from the template.
    Warning: This setting modifies the template in the ``sixdesk_directory``
    permanently. For more details see the htcondor API.


- **max_stage** *(str)*:

    Last stage to be run. All following stages are skipped.


- **python2** *(PathOrStr)*:

    Path to python to use with run_six.sh (python2 with requirements
    installed). ONLY THE PATH TO THE DIRECTORY OF THE python BINARY IS
    NEEDED! And it can't be an Anaconda Distribution. If ``None`` the
    system's ``python`` is used (SixDesk internally).

    default: ``None``


- **python3** *(PathOrStr)*:

    Path to python to use with sixdb (python3 with requirements
    installed).Defaults to the system's ``python3``.

    default: ``python3``


- **resubmit**:

    Resubmits to HTCondor if needed (i.e. in case it finds errors with the
    previous run).

    action: ``store_true``


- **sixdesk_directory** *(Path)*:

    Path to the directory of SixDesk. Defaults to the PRO-version on AFS.
    If you are using your own SixDesk Environment and it does not run,
    check the AutoSix doc.

    default: ``/afs/cern.ch/project/sixtrack/SixDesk_utilities/pro``


- **ssh** *(str)*:

    Run htcondor from this machine via ssh (needs access to the
    ``working_directory``)


- **stop_workspace_init**:

    Stops the workspace creation before initialization, so one can make
    manual changes.

    action: ``store_true``


- **unlock**:

    Forces unlocking of folders (if they have been locked by Sixdesk).

    action: ``store_true``


:author: jdilly

"""
import itertools
import logging
from pathlib import Path

import numpy as np
import tfs
from generic_parser import EntryPointParameters, entrypoint
from generic_parser.entry_datatypes import DictAsString

from pylhc_submitter.constants.autosix import (HEADER_BASEDIR, SIXENV_OPTIONAL, SIXENV_REQUIRED,
                                               AutoSixEnvironment)
from pylhc_submitter.constants.job_submitter import COLUMN_JOBID, JOBSUMMARY_FILE
from pylhc_submitter.submitter.mask import generate_jobdf_index
from pylhc_submitter.sixdesk_tools.stages import STAGE_ORDER, Stage
from pylhc_submitter.sixdesk_tools.utils import check_mask, is_locked
from pylhc_submitter.utils.iotools import (PathOrStr, keys_to_path, make_replace_entries_iterable,
                                           save_config)
from pylhc_submitter.utils.logging_tools import log_setup

LOG = logging.getLogger(__name__)


def get_params():
    params = EntryPointParameters()
    params.add_parameter(
        name="mask",
        type=Path,
        required=True,
        help="Path to the program's mask to use.",
    )
    params.add_parameter(
        name="working_directory",
        type=Path,
        required=True,
        help="Directory where data should be put into.",
    )
    params.add_parameter(
        name="replace_dict",
        help=(
            "Dict with keys of the strings to be replaced in the mask (required) "
            "as well as the mask_sixdeskenv and mask_sysenv files "
            "in the sixdesk_tools module. "
            f"Required fields are {', '.join(SIXENV_REQUIRED)}. "
            f"Optional fields are {', '.join(SIXENV_OPTIONAL)}. "
            "These keys can also be used in the mask if needed. "
            "The values of this dict are lists of values to replace "
            "these or single entries."
        ),
        type=DictAsString,
        required=True,
    )
    params.add_parameter(
        name="sixdesk_directory",
        type=Path,
        default=AutoSixEnvironment.sixdesk_directory,
        help="Path to the directory of SixDesk. Defaults to the PRO-version on AFS."
             " If you are using your own SixDesk Environment and it does not run, "
             " check the AutoSix doc.",
    )
    params.add_parameter(
        name="executable",
        default=AutoSixEnvironment.executable,
        type=PathOrStr,
        help="Path to executable or 'madx', 'python2', 'python3' "
             "to use the OMC default paths on AFS."
             "Defaults to the latest MADX-Binary on AFS.",
    )
    params.add_parameter(
        name="python2",
        default=AutoSixEnvironment.python2,
        type=PathOrStr,
        help=("Path to python to use with run_six.sh (python2 with requirements installed)."
              " ONLY THE PATH TO THE DIRECTORY OF THE python BINARY IS NEEDED!"
              " And it can't be an Anaconda Distribution."
              " If ``None`` the system's ``python`` is used (SixDesk internally)."),
    )
    params.add_parameter(
        name="python3",
        default=AutoSixEnvironment.python3,
        type=PathOrStr,
        help="Path to python to use with sixdb (python3 with requirements installed)."
             "Defaults to the system's ``python3``.",
    )
    params.add_parameter(
        name="jobid_mask",
        help="Mask-String to name jobs, with placeholders from the ``replace_dict`` keys.",
        type=str,
    )
    params.add_parameter(
        name="ssh",
        help="Run htcondor from this machine via ssh (needs access to the ``working_directory``)",
        type=str,
    )
    params.add_parameter(
        name="unlock",
        help="Forces unlocking of folders (if they have been locked by Sixdesk).",
        action="store_true",
    )
    params.add_parameter(
        name="apply_mad6t_hacks",
        help=(
            "Apply two hacks: Removes '<' in binary call and "
            "ignore the check for 'Twiss fail' in the submission file. "
            "This is hack needed in case this check greps the wrong lines, "
            "e.g. in madx-comments. USE WITH CARE!!"
        ),
        action="store_true",
    )
    params.add_parameter(
        name="stop_workspace_init",
        help=(
            "Stops the workspace creation before initialization,"
            " so one can make manual changes."
        ),
        action="store_true",
    )
    params.add_parameter(
        name="resubmit",
        help="Resubmits to HTCondor if needed "
             "(i.e. in case it finds errors with the previous run).",
        action="store_true",
    )
    params.add_parameter(
        name="da_turnstep",
        type=int,
        help="Step between turns used in DA-vs-Turns plot.",
        default=AutoSixEnvironment.da_turnstep,
    )
    params.add_parameter(
        name="max_stage",
        type=str,
        help="Last stage to be run. All following stages are skipped.",
    )
    params.add_parameter(
        name="max_materialize",
        type=int,
        help="Maximum jobs to be materialized in scheduler (per SixDesk Workspace!). "
             "Here: ``None`` leaves the settings as defined in the SixDesk "
             "htcondor_run_six.sub template and ``0`` removes it from the "
             "template. Warning: This setting modifies the template in the "
             "``sixdesk_directory`` permanently. For more details see the "
             "htcondor API.",
    )
    return params


@entrypoint(get_params(), strict=True)
def main(opt):
    """ Loop to create jobs from replace dict product matrix. """
    LOG.info("Starting autosix.")
    opt = _check_opts(opt)
    save_config(opt.working_directory, opt, "autosix")

    jobdf = _generate_jobs(
        opt.working_directory,
        opt.pop('jobid_mask'),  # not needed anymore
        **opt.pop('replace_dict')  # not needed anymore
    )
    env = AutoSixEnvironment(**opt)  # basically checks that everything is there

    for jobname, jobargs in jobdf.iterrows():
        run_job(jobname=jobname, jobargs=jobargs, env=env)


def run_job(jobname: str, jobargs: dict, env: AutoSixEnvironment):
    """Main submitting procedure for single job.

    Args:
        jobname (str): Name of the job/study
        env (DotDict): The ensemble of autosix settings as an ``AutoSixEnvironment`` object.
        jobargs(dict): All Key=Values needed to fill the mask!
    """
    if is_locked(jobname, env.working_directory, unlock=env.unlock):
        LOG.info(f"{jobname} is locked. Try 'unlock' flag if this causes errors.")

    Stage.run_all_stages(jobname, jobargs, env)


# Helper  ----------------------------------------------------------------------


def _check_opts(opt):
    opt = keys_to_path(opt, "mask", "working_directory", "executable")

    opt.mask_text = opt.mask.read_text()
    check_mask(opt.mask_text, opt.replace_dict)
    del opt.mask

    opt.replace_dict = make_replace_entries_iterable(opt.replace_dict)
    if opt.max_stage is not None and not isinstance(opt.max_stage, Stage):
        opt.max_stage = STAGE_ORDER[opt.max_stage]
    return opt


def get_jobs_and_values(jobid_mask, **kwargs):
    values_grid = np.array(list(itertools.product(*kwargs.values())), dtype=object)
    job_names = generate_jobdf_index(None, jobid_mask, kwargs.keys(), values_grid)
    return job_names, values_grid


def _generate_jobs(basedir, jobid_mask, **kwargs) -> tfs.TfsDataFrame:
    """ Generates product matrix for job-values and stores it as TfsDataFrame. """
    LOG.debug("Creating Jobs")
    job_names, values_grid = get_jobs_and_values(jobid_mask, **kwargs)
    job_df = tfs.TfsDataFrame(
        headers={HEADER_BASEDIR: basedir},
        index=job_names,
        columns=list(kwargs.keys()),
        data=values_grid,
    )
    tfs.write(basedir / JOBSUMMARY_FILE, job_df, save_index=COLUMN_JOBID)
    return job_df


if __name__ == "__main__":
    log_setup()
    main()



"""
Logging Tools
-------------

Tools that make the logging life easier.
"""
import logging
import sys


def log_setup():
    """ Set up a basic logger. """
    logging.basicConfig(
        stream=sys.stdout,
        level=logging.INFO,
        format="%(levelname)7s | %(message)s | %(name)s"
    )



"""
SixDesk Utilities
--------------------

Helper Utilities for Autosix.
"""
import logging
import subprocess
from pathlib import Path

from pylhc_submitter.constants.autosix import SIXDESKLOCKFILE, get_workspace_path
from pylhc_submitter.constants.external_paths import SIXDESK_UTILS
from pylhc_submitter.submitter.mask import find_named_variables_in_mask

LOG = logging.getLogger(__name__)


# Checks  ----------------------------------------------------------------------


def check_mask(mask_text: str, replace_args: dict):
    """ Checks validity/compatibility of the mask and replacement dict. """
    dict_keys = set(replace_args.keys())
    mask_keys = find_named_variables_in_mask(mask_text)
    not_in_dict = mask_keys - dict_keys

    if len(not_in_dict):
        raise KeyError(
            "The following keys in the mask were not found for replacement: "
            f"{str(not_in_dict).strip('{}')}"
        )


# Locks ------------------------------------------------------------------------


def is_locked(jobname: str, basedir: Path, unlock: bool = False):
    """ Checks for sixdesklock-files """
    workspace_path = get_workspace_path(jobname, basedir)
    locks = list(workspace_path.glob(f"**/{SIXDESKLOCKFILE}"))  # list() for repeated usage

    if locks:
        LOG.info("The following folders are locked:")
        for lock in locks:
            LOG.info(f"{str(lock.parent)}")

            with open(lock, "r") as f:
                txt = f.read()
            txt = txt.replace(str(SIXDESK_UTILS), "$SIXUTILS").strip("\n")
            if txt:
                LOG.debug(f" -> locked by: {txt}")

        if unlock:
            for lock in locks:
                LOG.debug(f"Removing lock {str(lock)}")
                lock.unlink()
            return False
        return True
    return False


# Commandline ------------------------------------------------------------------


def start_subprocess(command, cwd=None, ssh: str = None, check_log: str = None):
    if isinstance(command, str):
        command = [command]

    # convert Paths
    command = [str(c) if isinstance(c, Path) else c for c in command]

    if ssh:
        # Send command to remote machine
        command = " ".join(command)
        if cwd:
            command = f'cd "{cwd}" && {command}'
        LOG.debug(f"Executing command '{command}' on {ssh}")
        process = subprocess.Popen(
            ["ssh", ssh, command], shell=False, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=cwd,
        )

    else:
        # Execute command locally
        LOG.debug(f"Executing command '{' '.join(command)}'")
        process = subprocess.Popen(
            command, shell=False, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=cwd
        )

    # Log output
    for line in process.stdout:
        decoded = line.decode("utf-8").strip()
        if decoded:
            LOG.debug(decoded)
            if check_log is not None and check_log in decoded:
                raise OSError(
                    f"'{check_log}' found in last logging message. "
                    "Something went wrong with the last command. Check (debug-)log."
                )

    # Wait for finish and check result
    if process.wait() != 0:
        raise OSError("Something went wrong with the last command. Check (debug-)log.")



import pathlib

import setuptools

# The directory containing this file
MODULE_NAME = "pylhc_submitter"
TOPLEVEL_DIR = pathlib.Path(__file__).parent.absolute()
ABOUT_FILE = TOPLEVEL_DIR / MODULE_NAME / "__init__.py"
README = TOPLEVEL_DIR / "README.md"


def about_package(init_posixpath: pathlib.Path) -> dict:
    """
    Return package information defined with dunders in __init__.py as a dictionary, when
    provided with a PosixPath to the __init__.py file.
    """
    about_text: str = init_posixpath.read_text()
    return {
        entry.split(" = ")[0]: entry.split(" = ")[1].strip('"')
        for entry in about_text.strip().split("\n")
        if entry.startswith("__")
    }


ABOUT_PYLHC_SUBMITTER = about_package(ABOUT_FILE)

with README.open("r") as docs:
    long_description = docs.read()

# Dependencies for the module itself
DEPENDENCIES = [
    "numpy>=1.19",
    "pandas>=1.0",
    "htcondor>=8.9.2 ; sys_platform=='linux'",  # no bindings for macOS or windows on PyPI
    "tfs-pandas>=3.0",
    "generic-parser>=1.0.8",
    "scipy>=1.4.0",
    "matplotlib>=3.2.0",
]

EXTRA_DEPENDENCIES = {
    "test": [
        "pytest>=5.2",
        "pytest-cov>=2.7",
    ],
    "doc": ["sphinx", "sphinx_rtd_theme"],
}
EXTRA_DEPENDENCIES.update(
    {"all": [elem for list_ in EXTRA_DEPENDENCIES.values() for elem in list_]}
)


setuptools.setup(
    name=ABOUT_PYLHC_SUBMITTER["__title__"],
    version=ABOUT_PYLHC_SUBMITTER["__version__"],
    description=ABOUT_PYLHC_SUBMITTER["__description__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=ABOUT_PYLHC_SUBMITTER["__author__"],
    author_email=ABOUT_PYLHC_SUBMITTER["__author_email__"],
    url=ABOUT_PYLHC_SUBMITTER["__url__"],
    python_requires=">=3.7",
    license=ABOUT_PYLHC_SUBMITTER["__license__"],
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    packages=setuptools.find_packages(exclude=["tests*", "doc"]),
    include_package_data=True,
    install_requires=DEPENDENCIES,
    tests_require=EXTRA_DEPENDENCIES["test"],
    extras_require=EXTRA_DEPENDENCIES,
)



"""
HTCondor Utilities
------------------

This module provides functionality to create HTCondor jobs and submit them to ``HTCondor``.

``write_bash`` creates bash scripts executing either a python or madx script.
Takes as input `Dataframe`, job type, and optional additional commandline arguments for the script.
A shell script is created in each job directory in the dataframe.

``make_subfile`` takes the job dataframe and creates the **.sub** files required for submissions to
``HTCondor``. The **.sub** file will be put in the working directory. The maximum runtime of one
job can be specified, standard is 8h.
"""
import logging
import subprocess
from pathlib import Path
from typing import Any, Dict, List, Union

from pandas import DataFrame

from pylhc_submitter.constants.htcondor import (BASH_FILENAME, CMD_SUBMIT, HTCONDOR_JOBLIMIT,
                                                JOBFLAVOURS, NOTIFICATIONS, SHEBANG, SUBFILE)
from pylhc_submitter.constants.job_submitter import (COLUMN_DEST_DIRECTORY, COLUMN_JOB_DIRECTORY,
                                                     COLUMN_JOB_FILE, COLUMN_SHELL_SCRIPT,
                                                     EXECUTEABLEPATH, NON_PARAMETER_COLUMNS)
from pylhc_submitter.submitter import iotools
from pylhc_submitter.submitter.mask import is_mask_file
from pylhc_submitter.utils.environment import on_windows

try:
    import htcondor
except ImportError:  # will be handled by job_submitter
    class htcondor:
        """Dummy HTCondor module. To satisfy the typing. """
        Submit: Any = None


LOG = logging.getLogger(__name__)


# Subprocess Methods ###########################################################


def create_subfile_from_job(cwd: Path, submission: Union[str, htcondor.Submit]) -> Path:
    """
    Write file to submit to ``HTCondor``.
    
    Args:
        cwd (Path): working directory
        submission (str, htcondor.Submit): HTCondor submission definition (i.e. content of the file)

    Returns:
        Path: path to sub-file
    """
    subfile = cwd / SUBFILE
    LOG.debug(f"Writing sub-file '{str(subfile)}'.")
    with subfile.open("w") as f:
        f.write(str(submission))
    return subfile


def submit_jobfile(jobfile: Path, ssh: str) -> None:
    """Submit subfile to ``HTCondor`` via subprocess.
    
    Args:
        jobfile (Path): path to sub-file
        ssh (str): ssh target
    """
    proc_args = [CMD_SUBMIT, jobfile]
    if ssh:
        proc_args = ["ssh", ssh] + proc_args
    status = _start_subprocess(proc_args)
    if status:
        raise RuntimeError("Submit to HTCondor was not successful!")
    else:
        LOG.info("Jobs successfully submitted.")


def _start_subprocess(command: List[str]) -> int:
    """ Start subprocess and log output. 
    
    Args:
        command (List[str]): command to execute

    Returns:
        int: return code of the process
    """
    LOG.debug(f"Executing command '{command}'")
    process = subprocess.Popen(
        command, shell=False, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
    )
    for line in process.stdout:
        htc_line = line.decode("utf-8").strip()
        if htc_line:
            LOG.debug(f"{htc_line} (from HTCondor)")
    return process.wait()


# Job Creation #################################################################


def create_multijob_for_bashfiles(job_df: DataFrame, **kwargs) -> str:
    """
    Function to create an ``HTCondor`` submission content for all job-scripts, 
    i.e. bash-files, in the job_df.

    Keyword Args:
        output_dir (str): output directory that will be transferred. Defaults to ``None``.
        duration (str): max duration of the job. Needs to be one of the ``HTCondor`` Jobflavours.
            Defaults to ``workday``.
        group (str): force use of accounting group. Defaults to ``None``.
        retries (int): maximum amount of retries. Default to ``3``.
        notification (str): Notify under certain conditions. Defaults to ``error``.
        priority (int): Priority to order your jobs. Defaults to ``None``.

    Returns:
        str: HTCondor submission definition.
    """
    # Pre-defined HTCondor arguments for our jobs
    submit_dict = {
        "MyId": "htcondor",
        "universe": "vanilla",
        "arguments": "$(ClusterId) $(ProcId)",
        "output": Path("$(initialdir)", "$(MyId).$(ClusterId).$(ProcId).out"),
        "error": Path("$(initialdir)", "$(MyId).$(ClusterId).$(ProcId).err"),
        "log": Path("$(initialdir)", "$(MyId).$(ClusterId).$(ProcId).log"),
        "on_exit_remove": "(ExitBySignal == False) && (ExitCode == 0)",
        "requirements": "Machine =!= LastRemoteHost",
    }
    submit_dict.update(map_kwargs(kwargs))
    
    # Let the htcondor create the submit-file
    submission = htcondor.Submit(submit_dict)

    # add the multiple bash files
    scripts = [
        str(Path(*parts))
        for parts in zip(job_df[COLUMN_JOB_DIRECTORY], job_df[COLUMN_SHELL_SCRIPT])
    ]
    args = [",".join(parts) for parts in zip(scripts, job_df[COLUMN_JOB_DIRECTORY])]
    queueArgs = ["queue executable, initialdir from (", *args, ")"]

    # ugly but submission.setQArgs doesn't take string containing '\n':
    # submission.setQArgs("\n".join(queueArgs))  # doesn't work
    submission = str(submission) + "\n".join(queueArgs)
    LOG.debug(f"Created HTCondor subfile with content: \n{submission}")
    return submission


# Main functions ###############################################################


def make_subfile(cwd: Path, job_df: DataFrame, **kwargs) -> Path:
    """
    Creates submit-file for ``HTCondor``.
    For kwargs, see ``create_multijob_for_bashfiles``.

    Args:
        cwd (Path): working directory
        job_df (DataFrame): DataFrame containing all the job-information

    Returns:
        Path: path to the submit-file
    """
    job = create_multijob_for_bashfiles(job_df, **kwargs)
    return create_subfile_from_job(cwd, job)


def write_bash(
    job_df: DataFrame,
    output_dir: Path = None,
    executable: str = "madx",
    cmdline_arguments: dict = None,
    mask: Union[str, Path] = None,
) -> DataFrame:
    """
    Write the bash-files to be called by ``HTCondor``, which in turn call the executable.
    Takes as input `Dataframe`, job type, and optional additional commandline arguments for the script.
    A shell script is created in each job directory in the dataframe.

    Args:
        job_df (DataFrame): DataFrame containing all the job-information
        output_dir (str): output directory that will be transferred. Defaults to ``None``.
        executable (str): name of the executable. Defaults to ``madx``.
        cmdline_arguments (dict): additional commandline arguments for the executable
        mask (Union[str, Path]): string or path to the mask-file. Defaults to ``None``.

    Returns:
        DataFrame: The provided ``job_df`` but with added path to the scripts.
    """
    if len(job_df.index) > HTCONDOR_JOBLIMIT:
        raise AttributeError("Submitting too many jobs for HTCONDOR")

    exec_path = f"{str(EXECUTEABLEPATH.get(executable, executable))} " if executable else ''
    cmds = f" {' '.join([f'{param} {val}' for param, val in cmdline_arguments.items()])}" if cmdline_arguments else ''

    shell_scripts = [None] * len(job_df.index)
    for idx, (jobid, job) in enumerate(job_df.iterrows()):
        job_dir = Path(job[COLUMN_JOB_DIRECTORY])
        bash_file_name = f"{BASH_FILENAME}.{jobid}.{'bat' if on_windows() else 'sh'}"
        jobfile = job_dir / bash_file_name

        LOG.debug(f"Writing bash-file {idx:d} '{jobfile}'.")
        with open(jobfile, "w") as f:
            # Preparation ---
            if not on_windows():
                f.write(f"{SHEBANG}\n")
   
            if output_dir is not None:
                f.write(f"mkdir {str(output_dir)}\n")

            # The actual job execution ---
            f.write(exec_path)

            # Call the mask-file or the filled-template string
            if is_mask_file(mask):
                f.write(str(job_dir / job[COLUMN_JOB_FILE]))
            else:
                replace_columns = [column for column in job.index.tolist() if column not in NON_PARAMETER_COLUMNS]
                f.write(mask % dict(zip(replace_columns, job[replace_columns])))

            # Additional commands for the mask/string
            f.write(cmds)
            f.write("\n")

            # Manually copy output (if needed) ---
            dest_dir = job.get(COLUMN_DEST_DIRECTORY) 
            if output_dir and dest_dir and output_dir != dest_dir:
                if iotools.is_eos_uri(dest_dir):
                    # Note: eos-cp needs `/` at the end of both, source and target, dirs...
                    cp_command =  f'eos cp -r {_str_ending_with_slash(output_dir)} {_str_ending_with_slash(dest_dir)}'  
                else:
                    # ...but '/' at the end of source dir copies only the content on macOS.
                    cp_command =  f'cp -r {output_dir} {_str_ending_with_slash(dest_dir)}'  
                    
                f.write(f'{cp_command}\n')

        shell_scripts[idx] = bash_file_name

    job_df[COLUMN_SHELL_SCRIPT] = shell_scripts
    return job_df


def map_kwargs(add_dict: Dict[str, Any]) -> Dict[str, Any]:
    """
    Maps the kwargs for the job-file. 
    Some arguments have pre-defined choices and defaults, the remaining ones are just passed on.

    Args:
        add_dict (Dict[str, Any]): additional kwargs to add to the defaults.

    Returns:
        Dict[str, Any]: The mapped kwargs.
    """
    new = {}

    # Predefined mappings 
    htc_map = { # name: mapped_name, choices, default
        "duration": ("+JobFlavour", JOBFLAVOURS, "workday"),
        "output_dir": ("transfer_output_files", None, None),
        "accounting_group": ("+AccountingGroup", None, None),
        "max_retries": ("max_retries", None, 3),
        "notification": ("notification", NOTIFICATIONS, "error"),
    }
    for key, (mapped, choices, default) in htc_map.items():
        try:
            value = add_dict.pop(key)
        except KeyError:
            value = default  # could be `None`
        else:
            if choices is not None and value not in choices:
                raise TypeError(
                    f"{key} needs to be one of '{str(choices).strip('[]')}' but "
                    f"instead was '{value}'"
                )
        if value is not None:
            new[mapped] = _maybe_put_in_quotes(mapped, value)

    # Pass-Through Arguments
    LOG.debug(f"Remaining arguments to be added: '{str(add_dict).strip('{}'):s}'")
    new.update(add_dict)
    return new


# Helper #######################################################################

def _maybe_put_in_quotes(key: str, value: Any) -> Any:
    """ Put value in quoted strings if key starts with '+' """
    if key.startswith("+"):
        return f'"{value}"'
    return value


def _str_ending_with_slash(s: Union[Path, str]) -> str:
    """ Add a slash at the end of a path if not present. """
    s  = str(s)
    if s.endswith("/"):
        return s
    return f"{s}/"


# Script Mode ##################################################################


if __name__ == "__main__":
    raise EnvironmentError(f"{__file__} is not supposed to run as main.")



""" 
Job Submitter Runners
---------------------

Defines the methods to run the job-submitter, locally or on HTC.
"""
import logging
import multiprocessing
import subprocess
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Optional, Tuple

import pandas as pd
import tfs

from pylhc_submitter.constants.job_submitter import (COLUMN_DEST_DIRECTORY, COLUMN_JOB_DIRECTORY,
                                                     COLUMN_SHELL_SCRIPT)
from pylhc_submitter.submitter import htc_utils
from pylhc_submitter.submitter.iotools import is_eos_uri
from pylhc_submitter.utils.environment import on_windows

LOG = logging.getLogger(__name__)


@dataclass
class RunnerOpts:
    """ Options for running the submission. """
    working_directory: Path           # Path to the working directory (e.g. afs)
    jobflavour: Optional[str] = None  # HTCondor job flavour (lengths of the job)
    output_dir: Optional[str] = None  # Name of the output directory, where jobs store data
    ssh: Optional[str] = None         # SSH command
    dryrun: Optional[bool] = False    # Perform only a dry-run, i.e. do all but submit to HTC
    htc_arguments: Optional[Dict[str, Any]] = None  # Arguments to pass on to htc as keywords
    run_local: Optional[bool] = False # Run jobs locally
    num_processes: Optional[int] = 4  # Number of processes to run in parallel (locally)


def run_jobs(job_df: tfs.TfsDataFrame, opt: RunnerOpts) -> None:
    """Selects how to run the jobs.
    
    Args:
        job_df (tfs.TfsDataFrame): DataFrame containing all the job-information 
        opt (RunnerOpts): Parameters for the runner 
    """
    if opt.run_local: 
        run_local(job_df, opt)
    else:
        run_htc(job_df, opt)


def run_local(job_df: tfs.TfsDataFrame, opt: RunnerOpts) -> None:
    """Run all jobs locally.

    Args:
        job_df (tfs.TfsDataFrame): DataFrame containing all the job-information 
        opt (RunnerOpts): Parameters for the runner 
    """
    if opt.dryrun:
        LOG.info(f"Dry-run: Skipping local run.")
        return

    LOG.info(f"Running {len(job_df.index)} jobs locally in {opt.num_processes:d} processes.")
        
    pool = multiprocessing.Pool(processes=opt.num_processes)
    res = pool.map(_execute_shell, job_df.iterrows())
    if any(res):
        jobs_failed = [j for r, j in zip(res, job_df.index) if r]
        LOG.error(f"{len(jobs_failed)} of {len(job_df)} jobs have failed:\n {jobs_failed}")
        raise RuntimeError("At least one job has failed. Check output logs!")


def run_htc(job_df: tfs.TfsDataFrame, opt: RunnerOpts) -> None:
    """ Create submission file and submit the jobs to ``HTCondor``.

    Args:
        job_df (tfs.TfsDataFrame): DataFrame containing all the job-information
        opt (RunnerOpts): Parameters for the runner 
    """
    LOG.info(f"Submitting {len(job_df.index)} jobs on htcondor, flavour '{opt.jobflavour}'.")
    LOG.debug("Creating htcondor subfile.")

    subfile = htc_utils.make_subfile(
        opt.working_directory, job_df, 
        output_dir=opt.output_dir, 
        duration=opt.jobflavour, 
        **opt.htc_arguments
    )

    if opt.dryrun:
        LOG.info("Dry run: submission file created, but not submitting jobs to htcondor.")
        return

    LOG.debug("Submitting jobs to htcondor.")
    htc_utils.submit_jobfile(subfile, opt.ssh)


# Helper #######################################################################

def _execute_shell(df_row: Tuple[Any, pd.Series]) -> int:
    """ Execute the shell script. 
    
    Args:
        df_row (Tuple[Any, pd.Series]): Row in the job-dataframe as coming from `iterrows()`, 
                                        i.e. a tuple of (index, series)
    
    Returns:
        int: return code of the process
    """
    _, column = df_row
    cmd = [] if on_windows() else ["sh"]

    with Path(column[COLUMN_JOB_DIRECTORY], "log.tmp").open("w") as logfile:
        process = subprocess.Popen(
            cmd + [column[COLUMN_SHELL_SCRIPT]],
            shell=on_windows(),
            stdout=logfile,
            stderr=subprocess.STDOUT,
            cwd=column[COLUMN_JOB_DIRECTORY],
        )
    return process.wait()
